import struct
from typing import cast, Optional

from bxcommon.models.broadcast_message_type import BroadcastMessageType
from bxcommon.utils.stats import stats_format
from bxcommon.utils.stats.block_stat_event_type import BlockStatEventType
from bxcommon.utils.stats.block_statistics_service import block_stats
from bxcommon.utils.stats.stat_block_type import StatBlockType
from bxgateway import log_messages
from bxgateway.connections.abstract_relay_connection import AbstractRelayConnection
from bxgateway.services.block_processing_service import BlockProcessingService
from bxgateway.services.block_recovery_service import RecoveredTxsSource
from bxgateway.utils.errors.message_conversion_error import MessageConversionError
from bxutils import logging

logger = logging.get_logger(__name__)


class OntBlockProcessingService(BlockProcessingService):

    def __init__(self, node):
        super().__init__(node)
        self._node = cast("bxgateway.connections.ont.ont_gateway_node.OntGatewayNode", node)

    def process_block_broadcast(self, msg, connection: AbstractRelayConnection):
        is_consensus_msg, = struct.unpack_from("?", msg.blob()[8:9])
        if is_consensus_msg != self._node.opts.is_consensus:
            return
        super().process_block_broadcast(msg, connection)

    def retry_broadcast_recovered_blocks(self, connection):
        if self._node.block_recovery_service.recovered_blocks and self._node.opts.has_fully_updated_tx_service:
            for msg, recovered_txs_source in self._node.block_recovery_service.recovered_blocks:
                is_consensus_msg, = struct.unpack_from("?", msg[8:9])
                if is_consensus_msg and self._node.opts.is_consensus:
                    self._handle_decrypted_consensus_block(
                        msg,
                        connection,
                        recovered=True,
                        recovered_txs_source=recovered_txs_source
                    )
                else:
                    self._handle_decrypted_block(
                        msg,
                        connection,
                        recovered=True,
                        recovered_txs_source=recovered_txs_source
                    )

            self._node.block_recovery_service.clean_up_recovered_blocks()

    def _handle_decrypted_block(
        self,
        bx_block: memoryview,
        connection: AbstractRelayConnection,
        encrypted_block_hash_hex: Optional[str] = None,
        recovered: bool = False,
        recovered_txs_source: Optional[RecoveredTxsSource] = None
    ):
        is_consensus_msg, = struct.unpack_from("?", bx_block[8:9])
        if is_consensus_msg and self._node.opts.is_consensus:
            return self._handle_decrypted_consensus_block(
                bx_block,
                connection,
                encrypted_block_hash_hex,
                recovered,
                recovered_txs_source
            )
        else:
            return super()._handle_decrypted_block(
                bx_block,
                connection,
                encrypted_block_hash_hex,
                recovered,
                recovered_txs_source
            )

    def _handle_decrypted_consensus_block(
        self,
        bx_block: memoryview,
        connection: AbstractRelayConnection,
        encrypted_block_hash_hex: Optional[str] = None,
        recovered: bool = False,
        recovered_txs_source: Optional[RecoveredTxsSource] = None
    ):
        transaction_service = self._node.get_tx_service()

        if self._node.has_active_blockchain_peer() or self._node.remote_node_conn:
            try:
                block_message, block_info, unknown_sids, unknown_hashes = \
                    self._node.consensus_message_converter.bx_block_to_block(bx_block, transaction_service)
            except MessageConversionError as e:
                block_stats.add_block_event_by_block_hash(
                    e.msg_hash,
                    BlockStatEventType.BLOCK_CONVERSION_FAILED,
                    network_num=connection.network_num,
                    broadcast_type=BroadcastMessageType.CONSENSUS,
                    conversion_type=e.conversion_type.value
                )
                transaction_service.on_block_cleaned_up(e.msg_hash)
                connection.log_warning(log_messages.FAILED_TO_DECOMPRESS_BLOCK_ONT_CONSENSUS, e.msg_hash, e)
                return
        else:
            connection.log_warning(log_messages.LACK_BLOCKCHAIN_CONNECTION_ONT_CONSENSUS)
            return

        block_hash = block_info.block_hash
        all_sids = block_info.short_ids

        if encrypted_block_hash_hex is not None:
            block_stats.add_block_event_by_block_hash(block_hash,
                                                      BlockStatEventType.BLOCK_TO_ENC_BLOCK_MATCH,
                                                      matching_block_hash=encrypted_block_hash_hex,
                                                      matching_block_type=StatBlockType.ENCRYPTED.value,
                                                      network_num=connection.network_num,
                                                      broadcast_type=BroadcastMessageType.CONSENSUS)

        self.cancel_hold_timeout(block_hash, connection)

        if recovered:
            block_stats.add_block_event_by_block_hash(block_hash,
                                                      BlockStatEventType.BLOCK_RECOVERY_COMPLETED,
                                                      network_num=connection.network_num,
                                                      broadcast_type=BroadcastMessageType.CONSENSUS,
                                                      more_info=str(recovered_txs_source))

        if block_hash in self._node.blocks_seen.contents:
            block_stats.add_block_event_by_block_hash(block_hash, BlockStatEventType.BLOCK_DECOMPRESSED_IGNORE_SEEN,
                                                      start_date_time=block_info.start_datetime,
                                                      end_date_time=block_info.end_datetime,
                                                      network_num=connection.network_num,
                                                      broadcast_type=BroadcastMessageType.CONSENSUS,
                                                      prev_block_hash=block_info.prev_block_hash,
                                                      original_size=block_info.original_size,
                                                      compressed_size=block_info.compressed_size,
                                                      txs_count=block_info.txn_count,
                                                      blockchain_network=self._node.opts.blockchain_protocol,
                                                      blockchain_protocol=self._node.opts.blockchain_network,
                                                      matching_block_hash=block_info.compressed_block_hash,
                                                      matching_block_type=StatBlockType.COMPRESSED.value,
                                                      more_info=stats_format.duration(block_info.duration_ms))
            self._node.track_block_from_bdn_handling_ended(block_hash)
            transaction_service.track_seen_short_ids(block_hash, all_sids)
            connection.log_info(
                "Discarding duplicate consensus block {} from the BDN.",
                block_hash
            )
            return

        if not recovered:
            connection.log_info("Received consensus block {} from the BDN.", block_hash)
        else:
            connection.log_info("Successfully recovered consensus block {}.", block_hash)

        if block_message is not None:
            block_stats.add_block_event_by_block_hash(block_hash, BlockStatEventType.BLOCK_DECOMPRESSED_SUCCESS,
                                                      start_date_time=block_info.start_datetime,
                                                      end_date_time=block_info.end_datetime,
                                                      network_num=connection.network_num,
                                                      broadcast_type=BroadcastMessageType.CONSENSUS,
                                                      prev_block_hash=block_info.prev_block_hash,
                                                      original_size=block_info.original_size,
                                                      compressed_size=block_info.compressed_size,
                                                      txs_count=block_info.txn_count,
                                                      blockchain_network=self._node.opts.blockchain_protocol,
                                                      blockchain_protocol=self._node.opts.blockchain_network,
                                                      matching_block_hash=block_info.compressed_block_hash,
                                                      matching_block_type=StatBlockType.COMPRESSED.value,
                                                      peer=connection.peer_desc,
                                                      more_info="Consensus compression rate {}, Decompression time {}, "
                                                                "Queued behind {} blocks".format(
                                                          stats_format.percentage(block_info.compression_rate),
                                                          stats_format.duration(block_info.duration_ms),
                                                          self._node.block_queueing_service_manager.get_length_of_each_queuing_service_stats_format()))

            if recovered or block_hash in self._node.block_queueing_service_manager:
                self._node.block_queueing_service_manager.update_recovered_block(block_hash, block_message)
            else:
                self._node.block_queueing_service_manager.push(block_hash, block_message)

            self._node.block_recovery_service.cancel_recovery_for_block(block_hash)
            # self._node.blocks_seen.add(block_hash)
            transaction_service.track_seen_short_ids(block_hash, all_sids)
        else:
            if block_hash in self._node.block_queueing_service_manager and not recovered:
                connection.log_trace("Handling already queued consensus block again. Ignoring.")
                return

            self._node.block_recovery_service.add_block(bx_block, block_hash, unknown_sids, unknown_hashes)
            block_stats.add_block_event_by_block_hash(block_hash,
                                                      BlockStatEventType.BLOCK_DECOMPRESSED_WITH_UNKNOWN_TXS,
                                                      start_date_time=block_info.start_datetime,
                                                      end_date_time=block_info.end_datetime,
                                                      network_num=connection.network_num,
                                                      broadcast_type=BroadcastMessageType.CONSENSUS,
                                                      prev_block_hash=block_info.prev_block_hash,
                                                      original_size=block_info.original_size,
                                                      compressed_size=block_info.compressed_size,
                                                      txs_count=block_info.txn_count,
                                                      blockchain_network=self._node.opts.blockchain_protocol,
                                                      blockchain_protocol=self._node.opts.blockchain_network,
                                                      matching_block_hash=block_info.compressed_block_hash,
                                                      matching_block_type=StatBlockType.COMPRESSED.value,
                                                      more_info="{} sids, {} hashes".format(
                                                          len(unknown_sids), len(unknown_hashes)))

            connection.log_info("Consensus block {} requires short id recovery. Querying BDN...", block_hash)

            self.start_transaction_recovery(unknown_sids, unknown_hashes, block_hash, connection)
            if recovered:
                # should never happen –– this should not be called on blocks that have not recovered
                connection.log_error(log_messages.BLOCK_DECOMPRESSION_FAILURE_ONT_CONSENSUS,
                                     block_hash)
            else:
                self._node.block_queueing_service_manager.push(block_hash, waiting_for_recovery=True)



import importlib
from typing import Any, Optional, Union

from pympler import asizeof


class Sizer:

    def __init__(self, excluded_obj: Optional[Union[object, str]] = None) -> None:
        self.sizer = asizeof.Asizer()
        self.excluded = []

        if excluded_obj is not None:
            self.set_excluded_asizer(excluded_obj)

    def set_excluded_asizer(self, excluded_obj: object) -> Any:
        """
        :param excluded_obj: The string name or an instance of the class you want to exclude from memory calculation

        Sets the attributes of the singleton
        """
        if isinstance(excluded_obj, str):
            module_name, class_name = excluded_obj.rsplit(".", 1)
            class_instance = _str_to_class(module_name, class_name)
        else:
            class_instance = excluded_obj
        if class_instance:
            self.sizer.exclude_types(class_instance)
            self.excluded.append(str(excluded_obj))

    def __repr__(self) -> str:
        return "Sizer(sizer={0.sizer!r}, excluded={0.excluded!r})".format(self)

    def __str__(self) -> str:
        return "Sizer(sizer={0.sizer!s}, excluded={0.excluded!s})".format(self)


def _str_to_class(module_name, class_name):
    """
    Return a class instance from a string reference
    """
    class_instance = None
    try:
        module_instance = importlib.import_module(module_name)
        try:
            class_instance = getattr(module_instance, class_name)
        except AttributeError:
            pass
    except ImportError:
        pass
    return class_instance



import blxr_rlp as rlp
from bxgateway.messages.eth.protocol.eth_protocol_message import EthProtocolMessage
from bxgateway.messages.eth.protocol.eth_protocol_message_type import EthProtocolMessageType
from bxcommon.utils.blockchain_utils.eth import rlp_utils


class ReceiptsEthProtocolMessage(EthProtocolMessage):
    msg_type = EthProtocolMessageType.RECEIPTS

    fields = [("raw_data", rlp.sedes.raw)]

    def get_raw_data(self):
        return self.get_field_value("raw_data")

    def get_receipts_bytes(self):
        return rlp_utils.get_first_list_field_items_bytes(self._memory_view)



from typing import Optional

from bxcommon.utils.alarm_queue import AlarmQueue
from bxutils import logging

logger = logging.get_logger(__name__)


class IntervalMinimum:

    def __init__(self, interval_len_s: int, alarm_queue: AlarmQueue):
        self.current_minimum: int = 0

        self._interval_len_s = interval_len_s
        self._next_interval_minimum: Optional[int] = None
        alarm_queue.register_alarm(self._interval_len_s, self._on_interval_end)

    def add(self, new_value: int):
        next_interval_minimum = self._next_interval_minimum

        if next_interval_minimum is None or new_value < next_interval_minimum:
            self._next_interval_minimum = new_value

    def _on_interval_end(self):
        if self._next_interval_minimum is None:
            self.current_minimum = 0
        else:
            self.current_minimum = self._next_interval_minimum
        self._next_interval_minimum = None

        return self._interval_len_s



import struct
from typing import List, Optional

from bxcommon import constants
from bxcommon.messages.bloxroute import transactions_info_serializer
from bxcommon.messages.bloxroute.abstract_bloxroute_message import AbstractBloxrouteMessage
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.messages.bloxroute.txs_message import TxsMessage
from bxcommon.models.transaction_info import TransactionInfo
from bxcommon.utils import crypto
from bxcommon.utils.object_hash import Sha256Hash
from bxutils import logging

logger = logging.get_logger(__name__)


class CompressedBlockTxsMessage(AbstractBloxrouteMessage):
    MESSAGE_TYPE = BloxrouteMessageType.COMPRESSED_BLOCK_TXS
    """
    Message with tx details. Reply to GetTxsMessage.
    """

    _txs: Optional[List[TransactionInfo]] = None
    _network_num: Optional[int] = None
    _block_hash: Optional[Sha256Hash] = None

    def __init__(
        self,
        network_num: Optional[int] = None,
        block_hash: Optional[Sha256Hash] = None,
        txs: Optional[List[TransactionInfo]] = None,
        buf: Optional[bytearray] = None
    ) -> None:

        """
        Constructor. Expects list of transaction details or message bytes.
        :param network_num: Network number
        :param block_hash: Block hash
        :param txs: tuple with 3 values (tx short id, tx hash, tx contents)
        :param buf: message bytes
        """
        self.txs_info_offset = self.HEADER_LENGTH + constants.UL_INT_SIZE_IN_BYTES + crypto.SHA256_HASH_LEN

        if buf is None:
            assert network_num is not None
            assert block_hash is not None
            assert txs is not None
            buf = self._serialize(network_num, block_hash, txs)
        super().__init__(self.MESSAGE_TYPE, len(buf) - self.HEADER_LENGTH, buf)

        self._txs = None
        self._txs_count = None
        self._network_num = None
        self._block_hash = None

    def get_txs(self) -> List[TransactionInfo]:
        if self._txs is None:
            self._parse()

        txs = self._txs
        assert txs is not None
        return txs

    def get_txs_count(self) -> Optional[int]:
        if self._txs_count is None:
            off = self.HEADER_LENGTH + constants.UL_INT_SIZE_IN_BYTES + crypto.SHA256_HASH_LEN
            self._txs_count, = struct.unpack_from("<L", self.buf, off)

        return self._txs_count

    def block_hash(self) -> Sha256Hash:
        if self._block_hash is None:
            self._parse()

        block_hash = self._block_hash
        assert block_hash is not None
        return block_hash

    def network_num(self) -> int:
        if self._network_num is None:
            self._parse()

        network_num = self._network_num
        assert network_num is not None
        return network_num

    def to_txs_message(self) -> TxsMessage:
        txs_info_bytes = self.rawbytes()[self.txs_info_offset:]
        result_message_bytes = bytearray(TxsMessage.HEADER_LENGTH + len(txs_info_bytes))
        result_message_bytes[TxsMessage.HEADER_LENGTH:] = txs_info_bytes
        return TxsMessage(buf=result_message_bytes)

    def _serialize(self, network_num: int, block_hash: Sha256Hash, txs_details: List[TransactionInfo]):

        msg_size = (
            self.txs_info_offset
            + transactions_info_serializer.get_serialized_length(txs_details)
            + constants.CONTROL_FLAGS_LEN
        )

        buf = bytearray(msg_size)
        off = self.HEADER_LENGTH

        struct.pack_into("<L", buf, off, network_num)
        off += constants.UL_INT_SIZE_IN_BYTES

        buf[off:off + crypto.SHA256_HASH_LEN] = block_hash.binary
        off += crypto.SHA256_HASH_LEN

        transactions_info_serializer.serialize_transactions_info_to_buffer(txs_details, buf, off)

        return buf

    def _parse(self):
        off = self.HEADER_LENGTH

        self._network_num, = struct.unpack_from("<L", self.buf, off)
        off += constants.UL_INT_SIZE_IN_BYTES

        self._block_hash = Sha256Hash(self.buf[off: off + crypto.SHA256_HASH_LEN])
        off += crypto.SHA256_HASH_LEN

        txs, _ = transactions_info_serializer.deserialize_transactions_info_from_buffer(self.buf, off)

        self._txs = txs

    def __repr__(self):
        return "CompressedBlockTxsMessage<num_txs: {}>".format(len(self.get_txs()))

    def __iter__(self):
        for short_id in self.get_txs():
            yield short_id

    def __len__(self):
        return len(self.get_txs())



import datetime
import time
import unittest
from abc import abstractmethod
from typing import cast
from unittest.mock import MagicMock

from bxcommon import constants
from bxcommon.connections.connection_state import ConnectionState
from bxcommon.connections.connection_type import ConnectionType
from bxcommon.models.transaction_flag import TransactionFlag
from bxcommon.network.ip_endpoint import IpEndpoint
from bxcommon.rpc import rpc_constants
from bxcommon.test_utils import helpers
from bxcommon.test_utils.fixture import eth_fixtures
from bxcommon.test_utils.mocks.mock_connection import MockConnection
from bxcommon.test_utils.mocks.mock_socket_connection import MockSocketConnection
from bxgateway.connections.eth.eth_gateway_node import EthGatewayNode
from bxgateway.connections.eth.eth_node_connection import EthNodeConnection
from bxgateway.messages.eth.eth_normal_message_converter import EthNormalMessageConverter
from bxgateway.testing.mocks.mock_blockchain_connection import MockMessageConverter
from bxutils import constants as utils_constants
from bxcommon.models.bdn_account_model_base import BdnAccountModelBase
from bxcommon.models.bdn_service_model_base import BdnServiceModelBase, FeedServiceModelBase
from bxcommon.models.bdn_service_model_config_base import BdnServiceModelConfigBase, BdnFeedServiceModelConfigBase
from bxcommon.models.bdn_service_type import BdnServiceType
from bxcommon.models.blockchain_peer_info import BlockchainPeerInfo
from bxcommon.rpc.bx_json_rpc_request import BxJsonRpcRequest
from bxcommon.rpc.json_rpc_response import JsonRpcResponse
from bxcommon.rpc.rpc_request_type import RpcRequestType
from bxcommon.services.threaded_request_service import ThreadedRequestService
from bxcommon.test_utils.abstract_test_case import AbstractTestCase
from bxcommon.test_utils.helpers import async_test, AsyncMock
from bxcommon.utils import convert
from bxcommon.utils.object_hash import Sha256Hash
from bxgateway.gateway_opts import GatewayOpts
from bxgateway.rpc.gateway_status_details_level import GatewayStatusDetailsLevel
from bxgateway.rpc.requests import gateway_memory_rpc_request
from bxgateway.testing.mocks.mock_gateway_node import MockGatewayNode
from bxgateway.testing.mocks.mock_eth_ws_proxy_publisher import MockEthWsProxyPublisher
from bxgateway.utils.stats.gateway_bdn_performance_stats_service import \
    gateway_bdn_performance_stats_service


RAW_TRANSACTION_HEX = (
    "fffefdfc747800000000000000000000ab000000ffd59870844e5411f9e4043e654146b054bdcabe726a4bc4bd716049bfa"
    "54a690500000000000000000000000000000000000000000000000100000000f86b01847735940082520894a2f6090c0483"
    "d6e6ac90a9c23d42e461fee2ac5188016147191f13b0008025a0784537f9801331b707ceedd5388d318d86b0bb43c6f5b32"
    "b30c9df960f596b05a042fe22aa47f2ae80cbb2c9272df2f8975c96a8a99020d8ac19d4d4b0e0b5821901"
)

TRANSACTION_JSON = {'value': "", 'gas': "", 'gasPrice': "", 'from': '0xc305c901078781C232A2a521C2aF7980f8385ee9', 'to': '0xc305c901078781C232A2a521C2aF7980f8385ee9', 'data': '0x477a5c98'}

# TODO: this is different from relays? Not sure which is correct.
TRANSACTION_HASH = "74daac93cf5ff756e6fd5d10926a466533b4c901f64883e7f8b6b4a37cca808d"
ACCOUNT_ID = "bx_premium"


class AbstractGatewayRpcIntegrationTest(AbstractTestCase):
    def __init__(self, *args, **kwargs):
        # hack to avoid unit test discovery of this class
        super().__init__(*args, **kwargs)
        if self.__class__ != AbstractGatewayRpcIntegrationTest:
            # pylint: disable=no-value-for-parameter
            self.run = unittest.TestCase.run.__get__(self, self.__class__)
        else:
            self.run = lambda self, *args, **kwargs: None
        self._account_model = None

    # pyre-ignore async setup problems
    async def setUp(self) -> None:
        self.gateway_node = MockGatewayNode(self.get_gateway_opts())
        self.gateway_node.requester = ThreadedRequestService(
            "mock_thread_service",
            self.gateway_node.alarm_queue,
            constants.THREADED_HTTP_POOL_SLEEP_INTERVAL_S
        )
        self.gateway_node.requester.start()
        self.gateway_node.account_id = ACCOUNT_ID
        self.node_endpoint_1 = IpEndpoint("127.0.0.1", 7000)
        self.blockchain_connection_1 = EthNodeConnection(
            MockSocketConnection(1, node=self.gateway_node, ip_address="127.0.0.1", port=7000), cast(EthGatewayNode, self.gateway_node)
        )
        self.blockchain_connection_1.on_connection_established()

    @abstractmethod
    def get_gateway_opts(self) -> GatewayOpts:
        feed_service_model = FeedServiceModelBase(
            allow_filtering=True,
            available_fields=["all"]
        )
        base_feed_service_model = BdnFeedServiceModelConfigBase(
            expire_date="2999-01-01",
            feed=feed_service_model
        )
        self._account_model = BdnAccountModelBase(
            account_id="",
            logical_account_name="",
            certificate="",
            expire_date=utils_constants.DEFAULT_EXPIRATION_DATE.isoformat(),
            cloud_api=BdnServiceModelConfigBase(
                msg_quota=None,
                permit=BdnServiceModelBase(
                    service_type=BdnServiceType.PERMIT
                ),
                expire_date=utils_constants.DEFAULT_EXPIRATION_DATE.isoformat()
            ),
            new_transaction_streaming=base_feed_service_model,
            new_pending_transaction_streaming=base_feed_service_model,
            on_block_feed=base_feed_service_model,
            new_block_streaming=base_feed_service_model,
            transaction_state_feed=base_feed_service_model,
            blockchain_protocol="Ethereum",
            blockchain_network="Mainnet",
        )

    @abstractmethod
    async def request(self, req: BxJsonRpcRequest) -> JsonRpcResponse:
        pass

    @async_test
    async def test_blxr_tx(self):
        result = await self.request(BxJsonRpcRequest(
            "1",
            RpcRequestType.BLXR_TX,
            {
                rpc_constants.TRANSACTION_PARAMS_KEY: RAW_TRANSACTION_HEX,
                rpc_constants.STATUS_TRACK_PARAMS_KEY: "True"
            }
        ))
        self.assertEqual("1", result.id)
        self.assertIsNone(result.error)
        self.assertEqual(TRANSACTION_HASH, result.result["tx_hash"])

        self.assertEqual(1, len(self.gateway_node.broadcast_messages))
        self.assertEqual(
            Sha256Hash(convert.hex_to_bytes(TRANSACTION_HASH)),
            self.gateway_node.broadcast_messages[0][0].tx_hash()
        )

    @async_test
    async def test_blxr_tx_ethereum_berlin(self):
        self.gateway_node.message_converter = EthNormalMessageConverter()
        self.gateway_node.network_num = 5

        # legacy
        result = await self.request(BxJsonRpcRequest(
            "1",
            RpcRequestType.BLXR_TX,
            {
                rpc_constants.TRANSACTION_PARAMS_KEY: convert.bytes_to_hex(
                    eth_fixtures.LEGACY_TRANSACTION
                ),
                rpc_constants.STATUS_TRACK_PARAMS_KEY: "True"
            }
        ))
        self.assertEqual("1", result.id)
        self.assertIsNone(result.error)
        self.assertEqual(eth_fixtures.LEGACY_TRANSACTION_HASH, result.result["tx_hash"])

        self.assertEqual(1, len(self.gateway_node.broadcast_messages))
        self.assertEqual(
            Sha256Hash(convert.hex_to_bytes(eth_fixtures.LEGACY_TRANSACTION_HASH)),
            self.gateway_node.broadcast_messages[0][0].tx_hash()
        )

        # access list
        result = await self.request(BxJsonRpcRequest(
            "1",
            RpcRequestType.BLXR_TX,
            {
                rpc_constants.TRANSACTION_PARAMS_KEY: convert.bytes_to_hex(
                    eth_fixtures.ACL_TRANSACTION
                ),
                rpc_constants.STATUS_TRACK_PARAMS_KEY: "True"
            }
        ))
        self.assertEqual("1", result.id)
        self.assertIsNone(result.error)
        self.assertEqual(eth_fixtures.ACL_TRANSACTION_HASH, result.result["tx_hash"])

        self.assertEqual(2, len(self.gateway_node.broadcast_messages))
        self.assertEqual(
            Sha256Hash(convert.hex_to_bytes(eth_fixtures.ACL_TRANSACTION_HASH)),
            self.gateway_node.broadcast_messages[1][0].tx_hash()
        )

    @async_test
    async def test_blxr_tx_expired(self):
        self.gateway_node.account_model.is_account_valid = MagicMock(return_value=False)
        result = await self.request(BxJsonRpcRequest(
            "2",
            RpcRequestType.BLXR_TX,
            {
                rpc_constants.TRANSACTION_PARAMS_KEY: RAW_TRANSACTION_HEX,
            }
        ))
        self.assertEqual("2", result.id)
        self.assertEqual("Invalid Account ID", result.error.message)
        self.assertIsNone(result.result)
        self.assertIsNotNone(result.error)

    @async_test
    async def test_blxr_tx_quota_exceeded(self):
        self.gateway_node.quota_level = 100
        result = await self.request(BxJsonRpcRequest(
            "3",
            RpcRequestType.BLXR_TX,
            {
                rpc_constants.TRANSACTION_PARAMS_KEY: RAW_TRANSACTION_HEX,
            }
        ))
        self.assertEqual("3", result.id)
        self.assertEqual("Insufficient quota", result.error.message)
        self.assertIsNone(result.result)
        self.assertIsNotNone(result.error)

    @async_test
    async def test_blxr_tx_from_json(self):
        tx_json = {
            'from': '0xc165599b5e418bb9d8a19090696ea2403b2927ed',
            'gas': "0x5208",
            'gasPrice': "0xc1b759d70",
            'hash': '0xd569674ad9fcaaedcb6867b7896067b445d4a838316be4292c474df17bf4bd50',
            'input': '0x',
            'nonce': "0x14",
            'to': '0xdb5f0c1f4198bc6ffa98b35f7188f82740b8caf7',
            'value': "0x3e871b540c000",
            'v': '0x26',
            'r': '0x484bc950fb2d595500baa604774cb8d156a677198e087801936f38ca0b27049',
            's': '0x7ca82ef7b6938c2e96966dd940b186c9cdc0c7f42b2843adbc0751bb6e67a2d4'
        }

        result = await self.request(BxJsonRpcRequest(
            "1",
            RpcRequestType.BLXR_TX,
            {
                rpc_constants.TRANSACTION_JSON_PARAMS_KEY: tx_json,
            }
        ))

        self.assertEqual("1", result.id)
        self.assertIsNone(result.error)
        self.assertEqual("ad6f9332384194f80d8e49af8f093ad019b3f6b7173eb2956a46c9a0d8c4d03c", result.result["tx_hash"])

        self.assertEqual(1, len(self.gateway_node.broadcast_messages))
        self.assertEqual(
            Sha256Hash(convert.hex_to_bytes("ad6f9332384194f80d8e49af8f093ad019b3f6b7173eb2956a46c9a0d8c4d03c")),
            self.gateway_node.broadcast_messages[0][0].tx_hash()
        )

    @async_test
    async def test_blxr_eth_call(self):
        self.gateway_node.eth_ws_proxy_publisher = MockEthWsProxyPublisher(None, None, None, None)
        self.gateway_node.eth_ws_proxy_publisher.call_rpc = AsyncMock(
            return_value=JsonRpcResponse(request_id=1)
        )
        result = await self.request(BxJsonRpcRequest(
            "1",
            RpcRequestType.BLXR_ETH_CALL,
            {
                rpc_constants.TRANSACTION_JSON_PARAMS_KEY: TRANSACTION_JSON,
                rpc_constants.TAG_PARAMS_KEY: "latest"
            }
        ))
        self.gateway_node.eth_ws_proxy_publisher.call_rpc.mock.assert_called_with(
            "eth_call", [TRANSACTION_JSON, "latest"])
        self.assertIsNone(result.error)

    @async_test
    async def test_gateway_status(self):
        result_summary = await self.request(BxJsonRpcRequest(
            "2",
            RpcRequestType.GATEWAY_STATUS,
            None
        ))
        self.assertEqual("2", result_summary.id)
        self.assertIsNone(result_summary.error)
        self.assertEqual(constants.LOCALHOST, result_summary.result["ip_address"])
        self.assertEqual("NA", result_summary.result["continent"])
        self.assertEqual("United States", result_summary.result["country"])

        result_detailed = await self.request(BxJsonRpcRequest(
            "2.5",
            RpcRequestType.GATEWAY_STATUS,
            {
                rpc_constants.DETAILS_LEVEL_PARAMS_KEY: GatewayStatusDetailsLevel.DETAILED.name
            }
        ))
        self.assertEqual("2.5", result_detailed.id)
        self.assertIsNone(result_detailed.error)
        self.assertEqual(constants.LOCALHOST, result_detailed.result["summary"]["ip_address"])
        self.assertEqual("NA", result_detailed.result["summary"]["continent"])
        self.assertEqual("United States", result_detailed.result["summary"]["country"])

    @async_test
    async def test_stop(self):
        result = await self.request(BxJsonRpcRequest(
            "3",
            RpcRequestType.STOP,
            None
        ))
        self.assertEqual("3", result.id)
        self.assertIsNone(result.error)
        self.assertEqual({}, result.result)
        self.gateway_node.should_force_exit = True
        self.gateway_node.should_restart_on_high_memory = False

    @async_test
    async def test_memory(self):
        result = await self.request(BxJsonRpcRequest(
            "4",
            RpcRequestType.MEMORY,
            None
        ))
        self.assertEqual("4", result.id)
        self.assertIsNone(result.error)
        self.assertEqual(0, result.result[gateway_memory_rpc_request.TOTAL_CACHED_TX])
        self.assertEqual("0 bytes", result.result[gateway_memory_rpc_request.TOTAL_CACHED_TX_SIZE])

    @async_test
    async def test_peers(self):
        result = await self.request(BxJsonRpcRequest(
            "5",
            RpcRequestType.PEERS,
            None
        ))
        self.assertEqual("5", result.id)
        self.assertIsNone(result.error)
        self.assertEqual([], result.result)

    @async_test
    async def test_bdn_performance_single_node(self):
        self.gateway_node.connection_pool.add(
            20, self.node_endpoint_1.ip_address, self.node_endpoint_1.port, self.blockchain_connection_1
        )
        gateway_bdn_performance_stats_service.set_node(self.gateway_node)
        gateway_bdn_performance_stats_service.log_block_from_bdn()
        gateway_bdn_performance_stats_service.log_block_from_bdn()
        gateway_bdn_performance_stats_service.log_block_from_blockchain_node(self.node_endpoint_1)
        gateway_bdn_performance_stats_service.log_tx_from_bdn()
        gateway_bdn_performance_stats_service.log_tx_from_blockchain_node(self.node_endpoint_1)
        gateway_bdn_performance_stats_service.log_tx_from_blockchain_node(self.node_endpoint_1)
        gateway_bdn_performance_stats_service.close_interval_data()

        result = await self.request(BxJsonRpcRequest(
            "6",
            RpcRequestType.BDN_PERFORMANCE,
            None
        ))
        self.assertEqual("6", result.id)
        self.assertIsNone(result.error)
        node_stats = result.result[str(self.node_endpoint_1)]
        self.assertEqual("66.67%", node_stats["blocks_from_bdn_percentage"])
        self.assertEqual("33.33%", node_stats["transactions_from_bdn_percentage"])
        self.assertEqual(3, node_stats["total_blocks_seen"])

    @async_test
    async def test_bdn_performance_multi_node(self):
        self.gateway_node.connection_pool.add(
            20, self.node_endpoint_1.ip_address, self.node_endpoint_1.port, self.blockchain_connection_1
        )

        blockchain_connection_2 = EthNodeConnection(
            MockSocketConnection(1, node=self.gateway_node, ip_address="127.0.0.1", port=333), self.gateway_node)
        blockchain_connection_2.on_connection_established()
        self.gateway_node.mock_add_blockchain_peer(blockchain_connection_2)
        node_endpoint_2 = IpEndpoint("127.0.0.1", 333)
        self.gateway_node.connection_pool.add(
            21, node_endpoint_2.ip_address, node_endpoint_2.port, blockchain_connection_2
        )

        gateway_bdn_performance_stats_service.set_node(self.gateway_node)
        gateway_bdn_performance_stats_service.log_block_from_bdn()
        gateway_bdn_performance_stats_service.log_block_from_bdn()
        gateway_bdn_performance_stats_service.log_block_from_blockchain_node(self.node_endpoint_1)
        gateway_bdn_performance_stats_service.log_tx_from_bdn()
        gateway_bdn_performance_stats_service.log_tx_from_blockchain_node(self.node_endpoint_1)
        gateway_bdn_performance_stats_service.log_tx_from_blockchain_node(self.node_endpoint_1)
        gateway_bdn_performance_stats_service.close_interval_data()

        result = await self.request(BxJsonRpcRequest(
            "6",
            RpcRequestType.BDN_PERFORMANCE,
            None
        ))
        self.assertEqual("6", result.id)
        self.assertIsNone(result.error)
        node_stats = result.result[str(self.node_endpoint_1)]
        self.assertEqual("66.67%", node_stats["blocks_from_bdn_percentage"])
        self.assertEqual("33.33%", node_stats["transactions_from_bdn_percentage"])
        self.assertEqual(3, node_stats["total_blocks_seen"])

        node_stats_2 = result.result[str(node_endpoint_2)]
        self.assertEqual("100.00%", node_stats_2["blocks_from_bdn_percentage"])
        self.assertEqual("100.00%", node_stats_2["transactions_from_bdn_percentage"])
        self.assertEqual(3, node_stats_2["total_blocks_seen"])

    @async_test
    async def test_transaction_status(self):
        time.time = MagicMock(return_value=time.time())
        expected_assignment_time = datetime.datetime.fromtimestamp(time.time()).isoformat()

        short_id = 123
        transaction_hash = helpers.generate_object_hash()
        transaction_contents = helpers.generate_bytearray(250)

        tx_service = self.gateway_node.get_tx_service()
        transaction_key = tx_service.get_transaction_key(transaction_hash)
        tx_service.set_transaction_contents_by_key(transaction_key, transaction_contents)
        tx_service.assign_short_id_by_key(transaction_key, short_id)

        result = await self.request(BxJsonRpcRequest(
            "7",
            RpcRequestType.TX_STATUS,
            {"transaction_hash": convert.bytes_to_hex(transaction_hash.binary)}
        ))
        self.assertEqual("7", result.id)
        self.assertIsNone(result.error)
        self.assertEqual(
            {
                "status": "assigned short ID",
                "short_ids": [123],
                "assignment_time": expected_assignment_time
            },
            result.result
        )

    @async_test
    async def test_add_blockchain_peer(self):
        self.gateway_node.enqueue_connection = MagicMock()
        result = await self.request(BxJsonRpcRequest(
            "8",
            RpcRequestType.ADD_BLOCKCHAIN_PEER,
            {"peer": "enode://d76d7d11a822fab02836f8b0ea462205916253eb630935d15191fb6f9d218cd94a768fc5b3d5516b9ed5010a4765f95aea7124a39d0ab8aaf6fa3d57e21ef396@127.0.0.1:30302"}
        ))
        self.gateway_node.enqueue_connection.assert_called_once_with("127.0.0.1", 30302, ConnectionType.BLOCKCHAIN_NODE)
        self.assertIn(BlockchainPeerInfo("127.0.0.1", 30302), self.gateway_node.blockchain_peers)
        self.assertEqual("8", result.id)
        self.assertIsNone(result.error)
        self.assertEqual(
            {
                "new_peer": "127.0.0.1:30302"
            },
            result.result
        )

    @async_test
    async def test_remove_blockchain_peer(self):
        conn = MockConnection(MockSocketConnection(9, self.gateway_node, ip_address="127.0.0.1", port=30302), self.gateway_node)
        conn.mark_for_close = MagicMock()
        self.gateway_node.connection_pool.add(9, "127.0.0.1", 30302, conn)
        self.gateway_node.blockchain_peers.add(BlockchainPeerInfo("127.0.0.1", 30302, "d76d7d11a822fab02836f8b0ea462205916253eb630935d15191fb6f9d218cd94a768fc5b3d5516b9ed5010a4765f95aea7124a39d0ab8aaf6fa3d57e21ef396"))
        result = await self.request(BxJsonRpcRequest(
            "9",
            RpcRequestType.REMOVE_BLOCKCHAIN_PEER,
            {"peer": "enode://d76d7d11a822fab02836f8b0ea462205916253eb630935d15191fb6f9d218cd94a768fc5b3d5516b9ed5010a4765f95aea7124a39d0ab8aaf6fa3d57e21ef396@127.0.0.1:30302"}
        ))
        self.assertNotIn(BlockchainPeerInfo("127.0.0.1", 30302), self.gateway_node.blockchain_peers)
        conn.mark_for_close.assert_called_once_with(False)
        self.assertEqual("9", result.id)
        self.assertIsNone(result.error)
        self.assertEqual(
            {
                "removed_peer": "127.0.0.1:30302"
            },
            result.result
        )

    @async_test
    async def test_remove_blockchain_peer_not_in_pool(self):
        self.gateway_node.alarm_queue.register_alarm = MagicMock()
        self.gateway_node.blockchain_peers.add(BlockchainPeerInfo("127.0.0.1", 30302, "d76d7d11a822fab02836f8b0ea462205916253eb630935d15191fb6f9d218cd94a768fc5b3d5516b9ed5010a4765f95aea7124a39d0ab8aaf6fa3d57e21ef396"))
        result = await self.request(BxJsonRpcRequest(
            "9",
            RpcRequestType.REMOVE_BLOCKCHAIN_PEER,
            {"peer": "enode://d76d7d11a822fab02836f8b0ea462205916253eb630935d15191fb6f9d218cd94a768fc5b3d5516b9ed5010a4765f95aea7124a39d0ab8aaf6fa3d57e21ef396@127.0.0.1:30302"}
        ))
        self.assertNotIn(BlockchainPeerInfo("127.0.0.1", 30302), self.gateway_node.blockchain_peers)
        self.gateway_node.alarm_queue.register_alarm.assert_called()
        self.assertEqual("9", result.id)
        self.assertIsNone(result.error)
        self.assertEqual(
            {
                "removed_peer": "127.0.0.1:30302"
            },
            result.result
        )



from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.messages.bloxroute.v16.bdn_performance_stats_message_converter_v16 import \
    bdn_performance_stats_message_converter_v16
from bxcommon.messages.bloxroute.v17.message_converter_factory_v17 import message_converter_factory_v17
from bxcommon.messages.versioning.abstract_version_converter_factory import AbstractMessageConverterFactory
from bxcommon.messages.versioning.no_changes_message_converter import no_changes_message_converter


class _MessageConverterFactoryV16(AbstractMessageConverterFactory):
    _MESSAGE_CONVERTER_MAPPING = {
        # pylint: disable=protected-access
        **message_converter_factory_v17._MESSAGE_CONVERTER_MAPPING,
        BloxrouteMessageType.BDN_PERFORMANCE_STATS: bdn_performance_stats_message_converter_v16,
    }

    def get_message_converter(self, msg_type):
        if not msg_type:
            raise ValueError("msg_type is required.")

        if msg_type not in self._MESSAGE_CONVERTER_MAPPING:
            return no_changes_message_converter

        return self._MESSAGE_CONVERTER_MAPPING[msg_type]


message_converter_factory_v16 = _MessageConverterFactoryV16()



import struct
from typing import List, Optional

from bxcommon import constants
from bxcommon.messages.bloxroute.abstract_broadcast_message import AbstractBroadcastMessage
from bxcommon.utils import crypto
from bxcommon.utils.object_hash import Sha256Hash
from bxutils.logging.log_level import LogLevel


class AbstractCleanupMessage(AbstractBroadcastMessage):
    PAYLOAD_START_OFFSET = AbstractBroadcastMessage.HEADER_LENGTH + AbstractBroadcastMessage.PAYLOAD_LENGTH - \
        constants.CONTROL_FLAGS_LEN

    """
    Message with sids numbers for cleanup.
    """

    _tx_hashes: Optional[List[Sha256Hash]] = None
    _sids: Optional[List[int]] = None
    _sids_count: Optional[int] = None
    _tx_hashes_count: Optional[int] = None

    def __init__(
        self,
        message_hash: Optional[Sha256Hash] = None,
        network_num: Optional[int] = None,
        source_id: str = "",
        sids: Optional[List[int]] = None,
        tx_hashes: Optional[List[Sha256Hash]] = None,
        buf: Optional[bytearray] = None
    ) -> None:

        if buf is None:
            assert tx_hashes is not None and sids is not None
            # pylint: disable=invalid-name
            self.PAYLOAD_LENGTH = (
                AbstractBroadcastMessage.PAYLOAD_LENGTH
                + (constants.UL_INT_SIZE_IN_BYTES * 2)
                + (len(sids) * constants.UL_INT_SIZE_IN_BYTES)
                + (len(tx_hashes) * crypto.SHA256_HASH_LEN)
            )

        super(AbstractCleanupMessage, self).__init__(message_hash, network_num, source_id, buf)

        if buf is None:
            assert tx_hashes is not None and sids is not None
            sid_count = len(sids)
            hashes_count = len(tx_hashes)
            off = self.PAYLOAD_START_OFFSET

            struct.pack_into("<L", self.buf, off, sid_count)
            off += constants.UL_INT_SIZE_IN_BYTES

            for sid in sids:
                struct.pack_into("<L", self.buf, off, sid)
                off += constants.UL_INT_SIZE_IN_BYTES

            struct.pack_into("<L", self.buf, off, hashes_count)
            off += constants.UL_INT_SIZE_IN_BYTES

            for tx_hash in tx_hashes:
                self.buf[off:off + crypto.SHA256_HASH_LEN] = tx_hash
                off += crypto.SHA256_HASH_LEN

    def log_level(self) -> LogLevel:
        return LogLevel.DEBUG

    def short_ids(self) -> List[int]:
        if self._sids is None:
            self._parse()

        sids = self._sids
        assert sids is not None
        return sids

    def transaction_hashes(self) -> List[Sha256Hash]:
        if self._tx_hashes is None:
            self._parse()

        tx_hashes = self._tx_hashes
        assert tx_hashes is not None
        return tx_hashes

    def _parse(self) -> None:
        off = self.HEADER_LENGTH + AbstractBroadcastMessage.PAYLOAD_LENGTH - constants.CONTROL_FLAGS_LEN

        sids = []
        sids_count, = struct.unpack_from("<L", self.buf, off)
        self._sids_count = sids_count
        off += constants.UL_INT_SIZE_IN_BYTES

        for _ in range(sids_count):
            sid, = struct.unpack_from("<L", self.buf, off)
            off += constants.UL_INT_SIZE_IN_BYTES

            sids.append(sid)

        tx_hashes = []
        tx_hashes_count, = struct.unpack_from("<L", self.buf, off)
        self._tx_hashes_count = tx_hashes_count
        off += constants.UL_INT_SIZE_IN_BYTES

        for tx_hash in range(tx_hashes_count):
            tx_hash = Sha256Hash(self._memoryview[off:off + crypto.SHA256_HASH_LEN])
            off += crypto.SHA256_HASH_LEN

            tx_hashes.append(tx_hash)

        self._sids = sids
        self._tx_hashes = tx_hashes

    def __repr__(self) -> str:
        return "{} <message: {} :{}> <num_sids: {}> <num_tx_hashes: {}>".format(
            self.MESSAGE_TYPE.name,
            self.message_hash(),
            self._network_num,
            self._sids_count,
            self._tx_hashes_count
        )



import struct

from bxcommon.constants import MSG_NULL_BYTE
from bxcommon.exceptions import ChecksumError, PayloadLenError
from bxcommon.messages.abstract_message import AbstractMessage
from bxcommon.utils import crypto
from bxgateway.btc_constants import BTC_HDR_COMMON_OFF, BTC_HEADER_MINUS_CHECKSUM, BTC_MAGIC_NUMBERS
from bxutils import logging
from bxgateway import log_messages

logger = logging.get_logger(__name__)


class BtcMessage(AbstractMessage):
    MESSAGE_TYPE = b"btc"
    HEADER_LENGTH = BTC_HDR_COMMON_OFF

    def __init__(self, magic=None, command=None, payload_len=None, buf=None):
        self.buf = buf
        self._memoryview = memoryview(buf)

        magic_num = magic if magic not in BTC_MAGIC_NUMBERS else BTC_MAGIC_NUMBERS[magic]
        checksum = crypto.bitcoin_hash(self._memoryview[BTC_HDR_COMMON_OFF:payload_len + BTC_HDR_COMMON_OFF])

        off = 0
        struct.pack_into("<L12sL", buf, off, magic_num, command, payload_len)
        off += 20
        buf[off:off + 4] = checksum[0:4]

        self._magic = magic_num
        self._command = command
        self._payload_len = payload_len
        self._payload = None
        self._checksum = None

    # TODO: pull this out in to a message protocol
    @classmethod
    def unpack(cls, buf):
        magic, command, payload_length = struct.unpack_from("<L12sL", buf)
        checksum = buf[BTC_HEADER_MINUS_CHECKSUM:cls.HEADER_LENGTH]
        return command.rstrip(MSG_NULL_BYTE), magic, checksum, payload_length

    @classmethod
    def validate_payload(cls, buf, unpacked_args):
        command, _magic, checksum, payload_length = unpacked_args
        if payload_length != len(buf) - cls.HEADER_LENGTH:
            error_message = log_messages.PAYLOAD_LENGTH_MISMATCH.text.format(payload_length, len(buf))
            logger.error(error_message)
            raise PayloadLenError(error_message)
        ref_checksum = crypto.bitcoin_hash(buf[cls.HEADER_LENGTH:cls.HEADER_LENGTH + payload_length])[0:4]
        if checksum != ref_checksum:
            error_message = log_messages.PACKET_CHECKSUM_MISMATCH.text.format(checksum, ref_checksum, repr(buf))
            logger.error(error_message)
            raise ChecksumError(error_message, buf)

    @classmethod
    def initialize_class(cls, cls_type, buf, unpacked_args):
        instance = cls_type(buf=buf)
        return instance

    # END TODO

    def rawbytes(self) -> memoryview:
        """
        Returns a memoryview of the message
        """
        if self._payload_len is None:
            self._payload_len = struct.unpack_from("<L", self.buf, 16)[0]

        if self._payload_len + BTC_HDR_COMMON_OFF == len(self.buf):
            return self._memoryview
        else:
            return self._memoryview[0:self._payload_len + BTC_HDR_COMMON_OFF]

    def magic(self):
        if self._magic is None:
            self._magic, = struct.unpack_from("<L", self.buf)
        return self._magic

    def command(self):
        if self._command is None:
            self._command = struct.unpack_from("<12s", self.buf, 4)[0].rstrip(MSG_NULL_BYTE)
        return self._command

    def payload_len(self):
        if self._payload_len is None:
            self._payload_len, = struct.unpack_from("<L", self.buf, 16)
        return self._payload_len

    def checksum(self):
        if self._checksum is None:
            self._checksum = self.buf[BTC_HEADER_MINUS_CHECKSUM:BTC_HDR_COMMON_OFF]
        return self._checksum

    def payload(self):
        if self._payload is None:
            self._payload = self.buf[BTC_HDR_COMMON_OFF:self.payload_len() + BTC_HDR_COMMON_OFF]
        return self._payload

    def __eq__(self, other):
        """
        Expensive equality comparison. Use only for tests.
        """
        if not isinstance(other, BtcMessage):
            return False
        else:
            return self.rawbytes().tobytes() == other.rawbytes().tobytes()

    def __repr__(self):
        return "BtcMessage<command: {}, length: {}>".format(self.MESSAGE_TYPE, len(self.rawbytes()))



import logging
import os
import sys
import time
from asyncio import AbstractEventLoop
from logging import StreamHandler, FileHandler
from typing import Optional, List, Dict, Union, Iterable

from bxutils import constants
from bxutils import log_messages
from bxutils.encoding.json_encoder import EnhancedJSONEncoder
from bxutils.logging import log_level, LoggerConfig
from bxutils.logging.formatters import (
    JSONFormatter,
    CustomFormatter,
    FluentJSONFormatter,
    AbstractFormatter,
)
from bxutils.logging.handler_type import HandlerType
from bxutils.logging.log_format import LogFormat
from bxutils.logging.log_level import LogLevel

try:
    # TODO: remove try catch clause once the dependencies are installed
    from aiofluent.handler import FluentHandler
    # pylint: disable=ungrouped-imports
    from bxutils.logging.fluentd_logging_helpers import overflow_handler
except ImportError:
    FluentHandler = None
    fluentd_logging_helpers = None
    overflow_handler = None

logger = logging.getLogger(__name__)


def _get_handler_file(folder_path: str) -> FileHandler:
    assert folder_path is not None, "log folder path is missing"
    if not os.path.exists(folder_path):
        os.makedirs(folder_path)
    filename = os.path.join(
        folder_path,
        "{}{}.log".format(
            time.strftime("%Y-%m-%d-%H:%M:%S+0000-", time.gmtime()), str(os.getpid())
        ),
    )
    return FileHandler(filename)


def _get_handler_fluentd(
    fluentd_host: Optional[str],
    fluentd_tag_suffix: Optional[str],
    max_queue_size: int,
    loop: Optional[AbstractEventLoop] = None,
):
    assert fluentd_host is not None, "fluentd host name is missing"
    assert FluentHandler is not None, "fluentd handler is not installed"

    if ":" in fluentd_host:
        fluentd_host, fluentd_port = fluentd_host.split(":")
    else:
        fluentd_port = constants.FLUENTD_PORT
    if fluentd_tag_suffix:
        fluentd_tag = constants.FLUENTD_DEFAULT_TAG + "." + fluentd_tag_suffix
    else:
        fluentd_tag = constants.FLUENTD_DEFAULT_TAG
    return FluentHandler(
        fluentd_tag,
        host=fluentd_host,
        port=int(fluentd_port),
        buffer_overflow_handler=overflow_handler,
        nanosecond_precision=True,
        max_queue_size=max_queue_size,
        log_unhandled_exceptions=True,
        loop=loop,
        packer_kwargs={"default": EnhancedJSONEncoder().default},
    )


def _get_formatter(
    log_format: LogFormat, root_log_level: LogLevel, style: str, handler_type: HandlerType
):
    if log_format == LogFormat.PLAIN:
        if root_log_level <= LogLevel.DEBUG:
            formatter = CustomFormatter(
                fmt=constants.DEBUG_LOG_FORMAT_PATTERN,
                datefmt=constants.PLAIN_LOG_DATE_FORMAT_PATTERN,
                style=style,
            )
        else:
            formatter = CustomFormatter(
                fmt=constants.INFO_LOG_FORMAT_PATTERN,
                datefmt=constants.PLAIN_LOG_DATE_FORMAT_PATTERN,
                style=style,
            )
    elif handler_type == HandlerType.Fluent:
        formatter = FluentJSONFormatter(style=style)
    elif log_format == LogFormat.JSON:
        formatter = JSONFormatter(style=style)
    else:
        raise ValueError("LOG_FORMAT was not set correctly: {}".format(log_format))
    return formatter


def create_logger(
    global_logger_name: Optional[str],
    root_log_level: LogLevel = constants.DEFAULT_LOG_LEVEL,
    log_format: LogFormat = constants.DEFAULT_LOG_FORMAT,
    handler_type: HandlerType = HandlerType.Stream,
    flush_handlers: bool = True,
    folder_path: Optional[str] = None,
    style: str = "{",
    fluentd_host: Optional[str] = None,
    fluentd_tag_suffix: Optional[str] = None,
    max_queue_size=constants.FLUENTD_LOGGER_MAX_QUEUE_SIZE,
    handler_log_level: Optional[LogLevel] = None,
    loop: Optional[AbstractEventLoop] = None,
) -> None:
    """
    Installs a log configuration under the provided name.
    :param global_logger_name: log configuration name; None sets root logger
    :param root_log_level: the log level
    :param log_format: the logger format
    :param folder_path: optional file path (if specified - will write to the logs to files instead stdout)
    :param handler_type: enum log handler type
    :param flush_handlers: bool reset logging handlers
    :param style: the logger formatting style
    :param fluentd_host: fluentd host for fluent log handler
    :param fluentd_tag_suffix: optional fluentd tag suffix
    :param max_queue_size: the maximum size of the fluent logger queue
    :param handler_log_level: a specific log level for the handler itself (optional)
    :param loop: event loop optional, used in the fluentd handler
    """
    formatter = _get_formatter(log_format, root_log_level, style, handler_type)

    custom_logger = logging.getLogger(global_logger_name)

    if handler_type == HandlerType.File:
        assert folder_path is not None
        handler = _get_handler_file(folder_path)
    elif handler_type == HandlerType.Fluent:
        handler = _get_handler_fluentd(fluentd_host, fluentd_tag_suffix, max_queue_size, loop)
    else:
        handler = StreamHandler(sys.stdout)

    handler.setFormatter(formatter)
    if handler_log_level is not None:
        handler.setLevel(handler_log_level)

    custom_logger.propagate = False
    if custom_logger.hasHandlers() and flush_handlers:
        custom_logger.handlers = []
    custom_logger.addHandler(handler)
    custom_logger.setLevel(root_log_level)


def set_level(logger_names: List[Optional[str]], level: LogLevel) -> None:
    for logger_name in logger_names:
        logging.getLogger(logger_name).setLevel(level)


def set_log_levels(log_config: Dict[str, Union[LogLevel, str]]):
    # pyre-fixme[16]: `RootLogger` has no attribute `manager`.
    all_loggers = [logging.getLogger(name) for name in logging.root.manager.loggerDict]

    for logger_instance in all_loggers:
        for log_name, custom_log_level in log_config.items():
            if logger_instance.name.endswith(log_name):
                try:
                    if isinstance(custom_log_level, str):
                        custom_log_level = log_level.from_string(custom_log_level)
                    logging.getLogger(logger_instance.name).setLevel(custom_log_level)
                except (KeyError, AttributeError):
                    logger.error(log_messages.INVALID_LOG_LEVEL, log_name, custom_log_level)


def lazy_set_log_level(log_overrides) -> None:
    log_configs = {}
    for stats_logger_name in constants.STATS_LOGGER_NAMES:
        log_configs[stats_logger_name] = constants.DEFAULT_STATS_LOG_LEVEL
    log_configs.update(log_overrides)
    set_log_levels(log_configs)


def set_instance(instance: str) -> None:
    # TODO: change function signature
    AbstractFormatter.instance = instance


def str_to_log_options(value: str) -> Dict[str, LogLevel]:
    options = {}
    pairs = value.split(",")
    for pair in pairs:
        name, level = pair.split("=", 1)
        options[name] = log_level.from_string(level)
    return options


def setup_logging(
    log_format: LogFormat,
    default_log_level: LogLevel,
    default_logger_names: Iterable[str],
    log_level_overrides: Dict[str, LogLevel],
    root_log_level: LogLevel = LogLevel.WARNING,
    root_log_style: str = "{",
    enable_fluent_logger: bool = False,
    fluentd_host: Optional[str] = None,
    fluentd_queue_size: int = constants.FLUENTD_LOGGER_MAX_QUEUE_SIZE,
    fluentd_tag_suffix: Optional[str] = None,
    third_party_loggers: Optional[List[LoggerConfig]] = None,
    fluent_log_level: Optional[LogLevel] = None,
    stdout_log_level: Optional[LogLevel] = None,
    loop: Optional[AbstractEventLoop] = None,
) -> None:
    loggers_config = [LoggerConfig(None, root_log_style, root_log_level)]
    if third_party_loggers is not None:
        loggers_config.extend(third_party_loggers)

    for logger_config in loggers_config:
        if (
            logger_config.log_handler_type is None
            or logger_config.log_handler_type == HandlerType.Stream
        ):
            create_logger(
                logger_config.name,
                # pyre-fixme[6]: Expected `LogLevel` for 2nd param but got
                #  `Optional[LogLevel]`.
                root_log_level=logger_config.log_level
                if logger_config.log_level is not None
                else default_log_level,
                log_format=log_format,
                style=logger_config.style,
                handler_type=HandlerType.Stream,
                handler_log_level=stdout_log_level,
            )
        if (
            enable_fluent_logger
            and LoggerConfig.log_handler_type is None
            or LoggerConfig.log_handler_type == HandlerType.Fluent
        ):
            create_logger(
                logger_config.name,
                # pyre-fixme[6]: Expected `LogLevel` for 2nd param but got
                #  `Optional[LogLevel]`.
                root_log_level=logger_config.log_level
                if logger_config.log_level is not None
                else default_log_level,
                log_format=LogFormat.JSON,
                style=logger_config.style,
                handler_type=HandlerType.Fluent,
                flush_handlers=False,
                fluentd_host=fluentd_host,
                max_queue_size=fluentd_queue_size,
                fluentd_tag_suffix=fluentd_tag_suffix,
                handler_log_level=fluent_log_level,
                loop=loop,
            )
        elif enable_fluent_logger:
            print("Cannot Init fluentd logger")
    log_level_config = {}
    for logger_name in default_logger_names:
        log_level_config[logger_name] = default_log_level

    for stats_logger_name in constants.STATS_LOGGER_NAMES:
        log_level_config[stats_logger_name] = constants.DEFAULT_STATS_LOG_LEVEL

    log_level_config.update(log_level_overrides)
    set_log_levels(log_level_config)



import asyncio
from typing import TYPE_CHECKING, Dict, cast, Set, Union

import humps

from bxcommon import constants
from bxcommon.feed.feed import Feed
from bxcommon.feed.feed_source import FeedSource
from bxcommon.rpc import rpc_constants
from bxcommon.rpc.rpc_errors import RpcError
from bxcommon.utils.expiring_set import ExpiringSet
from bxgateway import gateway_constants
from bxgateway.feed.eth.eth_raw_block import EthRawBlock
from bxgateway.messages.eth.internal_eth_block_info import InternalEthBlockInfo
from bxgateway.services.eth.eth_block_queuing_service import EthBlockQueuingService
from bxutils import logging, utils

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxgateway.connections.eth.eth_gateway_node import EthGatewayNode

logger = logging.get_logger(__name__)

RETRIES_MAX_ATTEMPTS = 8


class TransactionReceiptsFeedEntry:
    receipt: Dict[str, str]

    def __init__(self, receipt: Dict):
        self.receipt = receipt

    def __eq__(self, other) -> bool:
        return (
            isinstance(other, TransactionReceiptsFeedEntry)
            and other.receipt == self.receipt
        )


class EthTransactionReceiptsFeed(Feed[TransactionReceiptsFeedEntry, Union[EthRawBlock, Dict]]):
    NAME = rpc_constants.ETH_TRANSACTION_RECEIPTS_FEED_NAME
    FIELDS = [
        "receipt",
        "receipt.block_hash",
        "receipt.block_number",
        "receipt.contract_address",
        "receipt.cumulative_gas_used",
        "receipt.from",
        "receipt.gas_used",
        "receipt.logs",
        "receipt.logs_bloom",
        "receipt.status",
        "receipt.to",
        "receipt.transaction_hash",
        "receipt.transaction_index"
    ]
    ALL_FIELDS = ["receipt"]
    VALID_SOURCES = {
        FeedSource.BLOCKCHAIN_SOCKET, FeedSource.BLOCKCHAIN_RPC, FeedSource.BDN_SOCKET
    }

    def __init__(self, node: "EthGatewayNode", network_num: int = constants.ALL_NETWORK_NUM, ) -> None:
        super().__init__(self.NAME, network_num)
        self.node = node
        self.last_block_number = 0
        self.published_blocks = ExpiringSet(
            node.alarm_queue, gateway_constants.MAX_BLOCK_CACHE_TIME_S, name="receipts_feed_published_blocks"
        )
        self.published_blocks_height = ExpiringSet(
            node.alarm_queue, gateway_constants.MAX_BLOCK_CACHE_TIME_S, name="receipts_feed_published_blocks_height"
        )
        self.blocks_confirmed_by_new_heads_notification = ExpiringSet(
            node.alarm_queue, gateway_constants.MAX_BLOCK_CACHE_TIME_S, name="receipts_feed_newHeads_confirmed_blocks"
        )

    def serialize(self, raw_message: Union[EthRawBlock, Dict]) -> TransactionReceiptsFeedEntry:
        # only receipts are serialized for publishing
        if isinstance(raw_message, Dict):
            return TransactionReceiptsFeedEntry(raw_message["result"])
        else:
            raise NotImplementedError

    def publish(self, raw_message: Union[EthRawBlock, Dict]) -> None:
        logger.trace(
            "attempting to publish message: {} for feed {}", raw_message, self.name
        )
        if raw_message.source not in self.VALID_SOURCES or self.subscriber_count() == 0:
            return
        if isinstance(raw_message, Dict):
            # transaction receipts published via parent publish method
            raise NotImplementedError

        block_hash = raw_message.block_hash
        block_number = raw_message.block_number
        block = raw_message.block

        # receipts won't be available until NewHeads feed notification
        if raw_message.source in [FeedSource.BLOCKCHAIN_RPC, FeedSource.BLOCKCHAIN_SOCKET]:
            self.blocks_confirmed_by_new_heads_notification.add(block_hash)

        if block_number < self.last_block_number - gateway_constants.MAX_BLOCK_BACKLOG_TO_PUBLISH:
            # published block is too far behind, ignore
            return

        if block_hash in self.published_blocks:
            # already published ignore
            return

        if block is None:
            block = cast(InternalEthBlockInfo, self.node.block_queuing_service_manager.get_block_data(block_hash))
            if block is None:
                return

        assert block is not None

        if block_hash not in self.blocks_confirmed_by_new_heads_notification:
            return

        if raw_message.source == FeedSource.BDN_SOCKET:
            block = block.to_new_block_msg()

        self.published_blocks.add(block_hash)
        self.published_blocks_height.add(block_number)

        if self.last_block_number and block_number > self.last_block_number + 1:
            # try to publish all intermediate blocks first
            missing_blocks = self._publish_blocks_from_queue(self.last_block_number + 1, block_number - 1)
            if missing_blocks:
                logger.info(
                    "Attempting to publish to feed block: {}, missing previous blocks {} ", block_number, missing_blocks
                )

        logger.debug("{} Attempting to fetch transaction receipts for block {}", self.name, block_hash)
        block_hash_str = block_hash.to_string(True)
        for tx in block.txns():
            asyncio.create_task(self._publish(tx.hash().to_string(True), block_hash_str))

        if block_number in self.published_blocks_height and block_number <= self.last_block_number:
            # possible fork, try to republish all later blocks
            _missing_blocks = self._publish_blocks_from_queue(block_number + 1, self.last_block_number)

        if block_number > self.last_block_number:
            self.last_block_number = block_number

    async def _publish(
        self,
        transaction_hash: str,
        block_hash: str,
        retry_count: int = 0
    ) -> None:
        response = None
        try:
            response = await self.node.eth_ws_proxy_publisher.call_rpc(
                rpc_constants.ETH_GET_TRANSACTION_RECEIPT_RPC_METHOD, [transaction_hash],
            )
        except RpcError as e:
            error_response = e.to_json()
            logger.warning(
                "Failed to fetch transaction receipt for {} in block {}: {}. Ceasing attempts.",
                transaction_hash, block_hash, error_response
            )

        assert response is not None

        if response.result is None:
            if retry_count == 0 or retry_count == RETRIES_MAX_ATTEMPTS:
                logger.debug(
                    "Failed to fetch transaction receipt for tx {} in block {}: not found. "
                    "Attempt: {}. Max attempts: {}.",
                    transaction_hash, block_hash, retry_count + 1, RETRIES_MAX_ATTEMPTS + 1
                )
            if retry_count < RETRIES_MAX_ATTEMPTS:
                sleep_time = utils.fibonacci(retry_count + 1) * 0.1
                await asyncio.sleep(sleep_time)
                asyncio.create_task(self._publish(transaction_hash, block_hash, retry_count + 1))
            return

        response.result = humps.decamelize(response.result)
        json_response = response.to_json()
        if json_response["result"]["block_hash"] != block_hash:
            return
        super().publish(json_response)

        if retry_count > 0:
            logger.debug(
                "Succeeded in fetching receipt for tx {} in block {} after {} attempts.",
                transaction_hash, block_hash, retry_count
            )

    def _publish_blocks_from_queue(self, start_block_height, end_block_height) -> Set[int]:
        missing_blocks = set()
        block_queuing_service = cast(
            EthBlockQueuingService,
            self.node.block_queuing_service_manager.get_designated_block_queuing_service()
        )
        if block_queuing_service is None:
            return missing_blocks

        for block_number in range(start_block_height, end_block_height):
            block_hash = block_queuing_service.accepted_block_hash_at_height.contents.get(block_number)
            if block_hash:
                block = self.node.block_queuing_service_manager.get_block_data(block_hash)
                if block is not None:
                    block_hash_str = block_hash.to_string(True)
                    for tx in block.txns():
                        asyncio.create_task(self._publish(tx.hash().to_string(True), block_hash_str))
            else:
                missing_blocks.add(block_number)
        return missing_blocks



import struct
from abc import ABCMeta

from bxcommon import constants
from bxcommon.constants import VERSION_NUM_LEN
from bxcommon.messages.bloxroute.abstract_bloxroute_message import AbstractBloxrouteMessage
from bxcommon.messages.bloxroute.tx_message import TxMessage
from bxcommon.messages.bloxroute.version_message import VersionMessage
from bxcommon.messages.versioning.nonversion_message_error import NonVersionMessageError
from bxcommon.utils.buffers.input_buffer import InputBuffer
from bxutils import log_messages
from bxutils import logging

logger = logging.get_logger(__name__)


class AbstractVersionManager:
    __metaclass__ = ABCMeta

    CURRENT_PROTOCOL_VERSION = 1
    MIN_SUPPORTED_PROTOCOL_VERSION = 1
    VERSION_MESSAGE_MAIN_LENGTH = VersionMessage.BASE_LENGTH

    def __init__(self) -> None:
        self.protocol_to_factory_mapping = {}
        self.protocol_to_converter_factory_mapping = {}
        self.version_message_command = ""

    def is_protocol_supported(self, protocol_version):
        return protocol_version >= self.MIN_SUPPORTED_PROTOCOL_VERSION

    def get_message_factory_for_version(self, protocol_version):
        """
        Returns message factory for provided protocol version

        :param protocol_version: protocol version
        :return: message factory
        """
        if not self.is_protocol_supported(protocol_version):
            raise ValueError("Protocol of version {} is not supported.".format(protocol_version))
        if protocol_version not in self.protocol_to_factory_mapping:
            logger.error(log_messages.PROTOCOL_VERSION_NOT_IN_FACTORY_MAPPING, protocol_version)
            raise NotImplementedError()

        return self.protocol_to_factory_mapping[protocol_version]

    def convert_message_to_older_version(self, convert_to_version: int, msg: AbstractBloxrouteMessage):
        """
        Converts message from current version to provided version

        :param convert_to_version: version to convert to
        :param msg: message
        :return: converted message
        """
        if not convert_to_version:
            raise ValueError("convert_to_version is required")

        if not msg:
            raise ValueError("msg is required")

        if convert_to_version not in self.protocol_to_converter_factory_mapping:
            raise ValueError("Conversion for version {} is not supported".format(convert_to_version))

        if isinstance(msg, TxMessage) and convert_to_version in msg.converted_message:
            converted_msg = msg.converted_message[convert_to_version]
        else:
            msg_converter = self._get_message_converter(convert_to_version, msg.msg_type())
            converted_msg = msg_converter.convert_to_older_version(msg)
            msg.converted_message[convert_to_version] = converted_msg
        return converted_msg

    def convert_message_from_older_version(self, convert_from_version, msg):
        """
        Converts message from older version to current version

        :param convert_from_version: version to convert from
        :param msg: message
        :return: converted message
        """
        if not convert_from_version:
            raise ValueError("convert_to_version is required")

        if not msg:
            raise ValueError("msg is required")

        if convert_from_version not in self.protocol_to_converter_factory_mapping:
            raise ValueError("Conversion for version {} is not supported".format(convert_from_version))

        msg_converter = self._get_message_converter(convert_from_version, msg.msg_type())
        return msg_converter.convert_from_older_version(msg)

    def convert_message_first_bytes_to_older_version(self, convert_to_version, msg_type, first_message_bytes):
        """
        Converts first message bytes from current version to provided version

        :param convert_to_version: version to convert to
        :param msg_type: message type
        :param first_message_bytes: message bytes
        :return: converted message bytes
        """

        if not convert_to_version:
            raise ValueError("convert_to_version is required")

        if not msg_type:
            raise ValueError("msg_type is required")

        if not first_message_bytes:
            raise ValueError("first_message_bytes is required")

        msg_converter = self._get_message_converter(convert_to_version, msg_type)

        return msg_converter.convert_first_bytes_to_older_version(first_message_bytes)

    def convert_message_first_bytes_from_older_version(self, convert_from_version, msg_type, first_message_bytes):
        """
        Converts first message bytes from older version to current version

        :param convert_from_version: version to convert from
        :param msg_type: message type
        :param first_message_bytes: message bytes
        :return: converted message bytes
        """

        if not convert_from_version:
            raise ValueError("convert_from_version is required")

        if not msg_type:
            raise ValueError("msg_type is required")

        if not first_message_bytes:
            raise ValueError("first_message_bytes is required")

        msg_converter = self._get_message_converter(convert_from_version, msg_type)

        return msg_converter.convert_first_bytes_from_older_version(first_message_bytes)

    def convert_message_last_bytes_to_older_version(self, convert_to_version, msg_type, message_last_bytes):
        """
        Converts message last bytes from current version to provided version

        :param convert_to_version: version to convert to
        :param msg_type: message type
        :param message_last_bytes: message bytes
        :return: converted message bytes
        """

        if not convert_to_version:
            raise ValueError("convert_to_version is required")

        if not msg_type:
            raise ValueError("msg_type is required")

        if not message_last_bytes:
            raise ValueError("first_message_bytes is required")

        msg_converter = self._get_message_converter(convert_to_version, msg_type)

        return msg_converter.convert_last_bytes_to_older_version(message_last_bytes)

    def convert_message_last_bytes_from_older_version(self, convert_from_version, msg_type, message_last_bytes):
        """
        Converts first message bytes from older version to current version

        :param convert_from_version: version to convert from
        :param msg_type: message type
        :param message_last_bytes: message bytes
        :return: converted message bytes
        """

        if not convert_from_version:
            raise ValueError("convert_from_version is required")

        if not msg_type:
            raise ValueError("msg_type is required")

        if not message_last_bytes:
            raise ValueError("first_message_bytes is required")

        msg_converter = self._get_message_converter(convert_from_version, msg_type)

        return msg_converter.convert_last_bytes_from_older_version(message_last_bytes)

    def get_message_size_change_to_older_version(self, convert_to_version, msg_type):
        """
        Returns the difference in size between current protocol version and older version

        :param convert_to_version: version to convert to
        :param msg_type: message type
        :return: size difference
        """

        if not convert_to_version:
            raise ValueError("convert_to_version is required")

        if not msg_type:
            raise ValueError("msg_type is required")

        msg_converter = self._get_message_converter(convert_to_version, msg_type)
        return msg_converter.get_message_size_change_to_older_version()

    def get_message_size_change_from_older_version(self, convert_from_version, msg_type):
        """
        Returns the difference in size between older version and current protocol version

        :param convert_from_version: version to convert from
        :param msg_type: message type
        :return: size difference
        """

        if not convert_from_version:
            raise ValueError("convert_to_version is required")

        if not msg_type:
            raise ValueError("msg_type is required")

        msg_converter = self._get_message_converter(convert_from_version, msg_type)

        return msg_converter.get_message_size_change_from_older_version()

    def get_connection_protocol_version(self, input_buffer: InputBuffer) -> int:
        if (
            input_buffer.length <
            (
                constants.STARTING_SEQUENCE_BYTES_LEN
                + constants.BX_HDR_COMMON_OFF
                + constants.VERSION_NUM_LEN
            )
        ):
            return self.CURRENT_PROTOCOL_VERSION

        header_buf = input_buffer.peek_message(VersionMessage.HEADER_LENGTH)
        if header_buf[:constants.STARTING_SEQUENCE_BYTES_LEN] == constants.STARTING_SEQUENCE_BYTES:
            command, payload_len = VersionMessage.unpack(header_buf)
            header_len = VersionMessage.HEADER_LENGTH
        else:
            command = bytearray(header_buf[:constants.MSG_TYPE_LEN])
            payload_len = constants.MSG_TYPE_LEN
            header_len = constants.BX_HDR_COMMON_OFF

        if command != self.version_message_command:
            if constants.HTTP_MESSAGE in command:
                raise NonVersionMessageError(
                    msg="Instead of a version hello message, we received an HTTP request: {} with payload length: {} "
                    "Ignoring and closing connection. "
                    .format(header_buf, payload_len),
                    is_known=True)

            if command in constants.BITCOIN_MESSAGES:
                raise NonVersionMessageError(
                    msg="Received some kind of bitcoin peering message: {}. "
                        "Ignoring and closing connection.".format(command),
                    is_known=True)

            logger.debug(
                "Received message of type {} instead of hello message. Use current version of protocol {}.",
                command,
                self.CURRENT_PROTOCOL_VERSION
            )
            return self.CURRENT_PROTOCOL_VERSION

        if payload_len < self.VERSION_MESSAGE_MAIN_LENGTH:
            return 1

        version_buf = input_buffer.get_slice(header_len, header_len + VERSION_NUM_LEN)
        version, = struct.unpack_from("<L", version_buf, 0)
        return version

    def _get_message_converter(self, version, msg_type):
        msg_converter_factory = self.protocol_to_converter_factory_mapping[version]
        return msg_converter_factory.get_message_converter(msg_type)



import logging
import typing
from dataclasses import dataclass
from typing import Optional, Union

from bxutils.logging.custom_logger import CustomLogger, CustomLogRecord
from bxutils.logging.handler_type import HandlerType
from bxutils.logging.log_level import LogLevel
from bxutils.logging.log_record_type import LogRecordType

logging.setLogRecordFactory(CustomLogRecord)  # pyre-ignore
logging.setLoggerClass(CustomLogger)


@dataclass()
class LoggerConfig:
    name: Optional[str]
    style: str
    log_level: Optional[LogLevel]
    log_handler_type: Optional[HandlerType] = None


def get_logger(name: Optional[Union[str, LogRecordType]] = None,
               parent_class_name: Optional[str] = None) -> CustomLogger:
    """
    Return a logger with the specified name, creating it if necessary.

    :param name: Name of the logger
    :param parent_class_name: The name of the class the logger is in
    :return: The logger
    """
    if isinstance(name, LogRecordType):
        name = name.value

    if parent_class_name:
        name = "{}.{}".format(parent_class_name, name)

    return typing.cast(CustomLogger, logging.getLogger(name))



import asyncio
import distutils
import json
import logging
import time
import operator
import argparse
import sys

from abc import abstractmethod
from collections import defaultdict, Counter
from distutils.util import strtobool
from enum import Enum
from typing import Union, Optional, Dict
from datetime import datetime

from bloxroute_cli.provider.ws_provider import WsProvider
from bxcommon.rpc.provider.abstract_ws_provider import AbstractWsProvider
from bxcommon.rpc.external.eth_ws_subscriber import EthWsSubscriber
from bloxroute_cli.provider.cloud_wss_provider import CloudWssProvider
from bxcommon.rpc.rpc_errors import RpcError

logger_summary = logging.getLogger("summary")
logger = logging.getLogger(f"{__name__}")
logger_block_result = logging.getLogger(f"{__name__}.block_result")


def setup_logger() -> None:
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(logging.Formatter('%(asctime)s - %(message)s'))
    logger.addHandler(handler)
    logging.getLogger("bxgateway").setLevel(logging.CRITICAL)
    logger.setLevel(logging.INFO)

    logger_summary.addHandler(logging.StreamHandler(sys.stdout))
    logger_summary.setLevel(logging.INFO)


class ResultCounter:
    def __init__(self, duration: int) -> None:
        self.notifications = dict()
        self.results = defaultdict(dict)
        self.summary = Counter()
        self.delta = Counter()
        self.duration = duration
        self.start = datetime.now()

    def log_results_summary(self, final: bool = True) -> None:
        current_time = datetime.now()
        blocks = sum(self.summary.values())
        node_types = ["Blxr", "Eth"]
        logger_summary.info("")
        if final:
            logger_summary.info(f"Start: {self.start}, End: {current_time}")
        else:
            logger_summary.info(f"Interval summary, Start: {self.start}, Current time: {current_time}")
        logger_summary.info(f"Blocks seen in duration: {len(self.results)}")
        logger_summary.info(f"Number of Blocks with results: {blocks}")
        for node_type in node_types:
            if node_type in self.summary:
                logger_summary.info(f"Number of results from {node_type} first: {self.summary[node_type]}")
        if blocks:
            ratio_from_gateway_first = self.summary.get("Blxr", 0) / blocks
            logger_summary.info(f"Percentage of results seen first from gateway: {ratio_from_gateway_first:.2%}")
            for node_type in node_types:
                if node_type in self.delta:
                    avg_delta = self.delta[node_type] / self.summary[node_type]
                    logger_summary.info(
                        f"Average time difference for results received first from {node_type} (ms): {avg_delta * 1000:.5f}"
                    )

    def log_notification(self, block_number: Union[str, int]) -> None:
        if block_number not in self.notifications:
            self.notifications[block_number] = time.time()

    def log_result_row(self, block_number: Union[str, int], name: str) -> None:
        if name in self.results[block_number]:
            logger.debug("Result was previously logged. Ignore block {} #{}.", name, block_number)
            return
        self.results[block_number][name] = time.time()
        row = self.results[block_number]
        if len(row) == 2:
            (winner, winner_res), (loser, loser_res) = sorted(row.items(), key=operator.itemgetter(1), reverse=False)
            delta = loser_res - winner_res
            if block_number in self.notifications:
                duration_winner = winner_res - self.notifications[block_number]
                duration_loser = loser_res - self.notifications[block_number]
            else:
                duration_winner, duration_loser = None, None
            logger_block_result.info(
                f"Block number: {block_number}: First response {winner}, "
                f"Duration from Block time {duration_winner * 1000:.5f} (ms). "
                f"{loser} Duration from Block time {duration_loser * 1000:.5f} (ms). "
                f"Delta {delta * 1000:.5f} (ms)"
            )
            self.summary[winner] += 1
            self.delta[winner] += delta
        else:
            logger_block_result.debug(
                f"Block number partial results: {block_number}: First: {name} "
                f"Duration first: {self.results[block_number][name] - self.notifications[block_number]:.5f} "
            )


class AbstractCallFeed:

    def __init__(self, call_params,  uri: str, name: str, counter: ResultCounter) -> None:
        self.call_params = call_params
        self.uri = uri
        self.name = name
        self.ws: Optional[AbstractWsProvider] = None
        self.subscription_id: Optional[str] = None
        self.counter: ResultCounter = counter

    async def __aenter__(self):
        await self.initialize()
        return self

    @abstractmethod
    async def initialize(self) -> None:
        pass


class CallFeed(AbstractCallFeed):

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.close()

    async def initialize(self) -> None:
        ws = await EthWsSubscriber(self.uri)
        assert ws.running
        self.subscription_id = await ws.subscribe("newHeads")
        logger.debug(f"Initialized subscription {self.name}: {self.subscription_id}")
        self.ws = ws
        asyncio.create_task(self.handle_block_notifications())

    async def close(self) -> None:
        ws = self.ws
        subscription_id = self.subscription_id
        if ws:
            if subscription_id:
                ws.unsubscribe(subscription_id)
            ws.close()

    async def handle_block_notifications(self, ) -> None:
        ws = self.ws
        subscription_id = self.subscription_id
        assert subscription_id is not None
        while ws and ws.running:
            next_notification = await ws.get_next_subscription_notification_by_id(subscription_id)
            block_header = next_notification.notification
            block_number = block_header["number"]
            if isinstance(block_number, str):
                block_number = int(block_number, 16)

            self.counter.log_notification(block_number)
            await self.process_block_notification(block_number)

    async def process_block_notification(self, block_number) -> None:
        tasks = []
        for call_params in self.call_params:
            tasks.append(
                asyncio.create_task(self.eth_call(block_number, **call_params))
            )

        await asyncio.wait(tasks, return_when=asyncio.ALL_COMPLETED)
        self.counter.log_result_row(block_number, self.name)

    async def eth_call(self, block_number, **kwargs) -> None:
        ws = self.ws
        name = kwargs.pop("name", None)
        command = kwargs.pop("method", "eth_call")
        address = kwargs.pop("address", "0x0000000000000000000000000000000000000000")
        pos = kwargs.pop("pos", "0x")
        tag = hex(block_number)

        if command == "eth_call":
            request_payload = [kwargs, tag]
        elif command in {"eth_getBalance", "eth_getCode", "eth_getTransactionCount"}:
            request_payload = [address, tag]
        elif command in {"eth_getStorageAt"}:
            request_payload = [address, pos, tag]
        elif command in {"eth_blockNumber"}:
            request_payload = []
        else:
            raise ValueError(f"Invalid EthCommand Option: {command}")

        assert ws is not None
        for i in range(10):
            try:
                _response = await ws.call_rpc(
                    command, request_payload
                )
                logger.debug(f"{self.name} #{block_number}: {_response.to_jsons()}")
                break
            except RpcError as e:
                if e.message == "header not found":
                    logger.debug(f"{e.message} retry {i}")
                else:
                    raise e
        return None


class BlxrCallFeed(CallFeed):
    connection_opts: Dict

    def __init__(
        self,
        call_params,
        uri: str,
        name: str,
        counter: ResultCounter,
    ) -> None:
        super(BlxrCallFeed, self).__init__(call_params, uri, name, counter)
        self.connection_opts = {}

    async def get_ws_connection(self):
        return await WsProvider(
            uri=self.uri,
            **self.connection_opts
        )

    async def initialize(self) -> None:
        ws = await self.get_ws_connection()
        assert ws.running

        sub_id = await ws.subscribe("ethOnBlock", options={
            "call_params": [
                {
                    "method": "eth_call",
                    **call_params
                }
                for call_params in self.call_params
                ]
            }
        )
        self.ws = ws
        self.subscription_id = sub_id
        logger.debug(f"Initialized subscription {self.name}: {self.subscription_id}")
        asyncio.create_task(self.wait_on_sub_id(sub_id))

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        ws = self.ws
        if ws is not None:
            await ws.close()

    async def wait_on_sub_id(self, sub_id: str) -> None:
        ws = self.ws
        while ws and ws.running:
            response = await ws.get_next_subscription_notification_by_id(sub_id)
            await self.handle_response(response)

    async def handle_response(self, response) -> None:
        block_number = response.notification["blockHeight"]
        logger.debug(f"{self.name} #{block_number}: {response.notification['response']}")
        self.counter.log_notification(block_number)
        if response.notification["name"] == "TaskDisabledEvent":
            logger.error(f"Invalid subscription task, task was disabled {response.notification}")
        if response.notification["name"] == "TaskCompletedEvent":
            self.counter.log_result_row(block_number, self.name)


class BlxrCloudCallFeed(BlxrCallFeed):
    def __init__(
        self,
        call_params,
        uri: str,
        name: str,
        counter: ResultCounter,
        ca_url: Optional[str] = None,
        ssl_dir: Optional[str] = None
    ) -> None:
        super(BlxrCloudCallFeed, self).__init__(call_params, uri, name, counter)
        if ca_url:
            self.connection_opts["ca_url"] = ca_url
        if ssl_dir:
            self.connection_opts["ssl_dir"] = ssl_dir

    async def get_ws_connection(self):
        return await CloudWssProvider(
            ws_uri=self.uri,
            **self.connection_opts
        )


def _get_feed_connection(
    eth_call_params, conn_type, conn_uri, conn_name, counter, ca_url=None, ssl_dir=None
) -> AbstractCallFeed:
    if conn_type == "Node":
        conn = CallFeed(eth_call_params, conn_uri, conn_name, counter)
    elif conn_type == "Local":
        conn = BlxrCallFeed(eth_call_params, conn_uri, conn_name, counter)
    elif conn_type == "Cloud":
        conn = BlxrCloudCallFeed(eth_call_params, conn_uri, conn_name, counter, ca_url=ca_url, ssl_dir=ssl_dir)
    else:
        raise ValueError
    return conn


async def log_counter_stats(counter: ResultCounter, summary_interval: int = 0) -> None:
    while summary_interval > 0:
        await asyncio.sleep(summary_interval)
        counter.log_results_summary(final=False)


async def main(opts: argparse.Namespace, counter: ResultCounter) -> None:
    logger.info(f"Started feed comparison. Duration {opts.duration}s.")
    conn_a = _get_feed_connection(
        opts.eth_call_params, "Node", opts.eth, "Eth", counter,
    )
    conn_b = _get_feed_connection(
        opts.eth_call_params,
        opts.gateway_type, opts.gateway, "Blxr", counter,
        ssl_dir=opts.gateway_ssl_dir, ca_url=opts.gateway_ca_cert_url
    )
    if opts.use_eth_node:
        await conn_a.initialize()
    await conn_b.initialize()
    if opts.summary_interval:
        asyncio.create_task(log_counter_stats(counter, summary_interval=opts.summary_interval))

    await asyncio.sleep(opts.duration)
    counter.log_results_summary(final=True)


def get_opts(argv=None) -> argparse.Namespace:
    DEFAULT_ETH_NODE_URI = "ws://127.0.0.1:8546"
    DEFAULT_GATEWAY_NODE_URI = "wss://eth.feed.blxrbdn.com:28333"

    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument("--eth",
                            help=f"Eth node uri. Default: {DEFAULT_ETH_NODE_URI}",
                            type=str,
                            default=DEFAULT_ETH_NODE_URI)
    arg_parser.add_argument("--use-eth-node",
                            help=f"Default: True, set to false to test only Gateway performance",
                            type=strtobool,
                            default=True)
    arg_parser.add_argument("--gateway",
                            help=f"Gateway node uri. Default: {DEFAULT_GATEWAY_NODE_URI}",
                            type=str,
                            default=DEFAULT_GATEWAY_NODE_URI)
    arg_parser.add_argument("--use-cloud-api",
                            help="Default True(CloudApi) False for local Gateway node",
                            type=strtobool,
                            default=True)
    arg_parser.add_argument("--gateway-ssl-dir",
                            help="for cloud connection only path to ssl certificate",
                            type=str,
                            )
    arg_parser.add_argument("--gateway-ca_cert_url",
                            help="for bdn cloud only url for ca cert",
                            default="https://certificates.blxrbdn.com/ca/ca_cert.pem",
                            type=str,
                            )
    arg_parser.add_argument("--duration",
                            help="test duration in seconds",
                            type=int,
                            default=600
                            )
    arg_parser.add_argument("--summary-interval",
                            help="display summary every N seconds",
                            type=int,
                            default=0
                            ),
    arg_parser.add_argument("-v", "--verbose",
                            help="display per block results",
                            action="store_true",
                            default=False
                            ),
    arg_parser.add_argument("-d", "--debug",
                            help="log debug information",
                            action="store_true",
                            default=False
                            ),

    arg_parser.add_argument("--call-params-file",
                            help="Json File that contains the ETH call settings, please see example file",
                            type=str,
                            default="on_block_feed_call_params_example.json"
                            )

    opts = arg_parser.parse_args(argv)
    with open(opts.call_params_file, "rb+") as f:
        opts.eth_call_params = json.loads(f.read())
    if opts.debug:
        log_level = logging.DEBUG
        logger.setLevel(logging.DEBUG)
    else:
        log_level = logging.INFO
    if opts.verbose:
        logger_block_result.setLevel(log_level)

    if opts.use_cloud_api:
        opts.gateway_type = "Cloud"
    else:
        opts.gateway_type = "Local"
    return opts


if __name__ == "__main__":
    setup_logger()
    opts = get_opts()
    result_counter = ResultCounter(
        duration=opts.duration
    )
    try:
        asyncio.run(
            main(opts, result_counter)
        )
    except KeyboardInterrupt:
        logger.error("Exit requested. Shutting down")
        result_counter.log_results_summary()



from typing import Optional

from bxcommon import constants
from bxcommon.messages.bloxroute.abstract_bloxroute_message import AbstractBloxrouteMessage
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxutils.logging import LogLevel


class RefreshBlockchainNetworkMessage(AbstractBloxrouteMessage):
    MESSAGE_TYPE = BloxrouteMessageType.REFRESH_BLOCKCHAIN_NETWORK
    """
    Request sdn to refresh blockchain network model
    """

    def __init__(self, buf: Optional[bytearray] = None) -> None:
        if buf is None:
            buf = bytearray(self.HEADER_LENGTH + constants.CONTROL_FLAGS_LEN)
            self.buf = buf

        super(RefreshBlockchainNetworkMessage, self).__init__(self.MESSAGE_TYPE, constants.CONTROL_FLAGS_LEN, buf)

    def log_level(self):
        return LogLevel.DEBUG

    def __repr__(self):
        return "RefreshBlockchainNetworkMessage<>"



from collections import deque
from typing import List, Deque


class RunningAverage:

    def __init__(self, max_count: int) -> None:
        self.max_count: int = max_count
        self.average: float = 0
        self.values: Deque[float] = deque()

        self.max_next_multiplier = 1 / self.max_count

    def add_value(self, value: float) -> None:
        self.values.append(value)
        if len(self.values) <= self.max_count:
            previous_multiplier = (len(self.values) - 1) / len(self.values)
            next_multiplier = 1 / len(self.values)
            self.average = self.average * previous_multiplier + value * next_multiplier
        else:
            oldest_item = self.values.popleft()
            self.average -= oldest_item * self.max_next_multiplier
            self.average += value * self.max_next_multiplier



from typing import Dict, Any, TYPE_CHECKING, List, Type

from bxcommon.utils.stats.statistics_service import StatisticsService, StatsIntervalData
from bxcommon.rpc import rpc_constants
from bxgateway import eth_constants


from bxutils import logging
from bxutils.logging import LogRecordType


if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode


class EthOnBlockFeedStatInterval(StatsIntervalData):
    subscriber_task_count: List[int]
    subscriber_task_duration: List[float]

    def __init__(self):
        super().__init__()
        self.subscriber_task_count = [0]
        self.subscriber_task_duration = [0]


class EthOnBlockFeedStatsService(
    StatisticsService[EthOnBlockFeedStatInterval, "AbstractGatewayNode"]
):
    def __init__(
        self,
        interval: int = eth_constants.ETH_ON_BLOCK_FEED_STATS_INTERVAL_S,
        look_back: int = eth_constants.ETH_ON_BLOCK_FEED_STATS_LOOKBACK,
    ) -> None:
        super().__init__(
            "EthCallFeedStats",
            interval,
            look_back,
            reset=True,
            stat_logger=logging.get_logger(LogRecordType.EthOnBlockFeedStats),
        )
        self.feed_name = rpc_constants.ETH_ON_BLOCK_FEED_NAME

    def get_interval_data_class(self) -> Type[EthOnBlockFeedStatInterval]:
        return EthOnBlockFeedStatInterval

    def get_info(self) -> Dict[str, Any]:
        interval_data = self.interval_data

        node = self.node
        assert node is not None
        feeds = node.feed_manager.feeds
        subscriber_count = 0
        if feeds and self.feed_name in feeds:
            feed = feeds[self.feed_name]
            subscriber_count = len(feed.subscribers)

        return {
            "subscriber_count": subscriber_count,
            "total_calls": sum(interval_data.subscriber_task_count),
            "max_calls_per_subscriber": max(interval_data.subscriber_task_count, default=None),
            "max_duration": max(interval_data.subscriber_task_duration, default=None),
        }

    def log_subscriber_tasks(self, tasks_count: int, duration_s: float) -> None:
        interval_data = self.interval_data
        interval_data.subscriber_task_count.append(tasks_count)
        interval_data.subscriber_task_duration.append(duration_s)


eth_on_block_feed_stats_service = EthOnBlockFeedStatsService()



import json
from typing import Any, Optional, Dict, Union

import humps

from bxcommon.rpc import rpc_constants
from bxcommon.rpc.rpc_errors import RpcError
from bxutils.encoding import json_encoder
from bxutils import utils
from bxutils.encoding.json_encoder import Case


class JsonRpcResponse:
    id: Optional[str]
    result: Optional[Any]
    error: Optional[RpcError]

    def __init__(
        self,
        request_id: Optional[str],
        result: Optional[Any] = None,
        error: Optional[RpcError] = None,
    ) -> None:
        if (result is not None) and (error is not None):
            raise ValueError(
                "Cannot instantiate a JsonRpcResponse with both an error and a result."
            )

        self.id = request_id
        self.result = result
        self.error = error
        self.json_rpc_version = rpc_constants.JSON_RPC_VERSION

    def __str__(self) -> str:
        return f"JsonRpcResponse<{self.to_jsons()}>"

    def __eq__(self, o: object) -> bool:
        if isinstance(o, JsonRpcResponse):
            return (
                self.id == o.id
                and self.result == o.result
                and self.error == o.error
                and self.json_rpc_version == o.json_rpc_version
            )
        else:
            return False

    def to_json(self, case: Case = Case.SNAKE) -> Dict[str, Any]:
        fields = {
            rpc_constants.JSON_RPC_VERSION_FIELD: self.json_rpc_version,
            rpc_constants.JSON_RPC_REQUEST_ID: self.id,
        }
        result = self.result
        if result is not None:
            if case == Case.CAMEL and isinstance(result, dict):
                fields[rpc_constants.JSON_RPC_RESULT] = humps.camelize(
                    result
                )
            else:
                fields[rpc_constants.JSON_RPC_RESULT] = result

        error = self.error
        if error is not None:
            fields[rpc_constants.JSON_RPC_ERROR] = error.to_json()
        return fields

    def to_jsons(self, case: Case = Case.SNAKE) -> str:
        return json_encoder.to_json(self.to_json(case))

    @classmethod
    def from_json(cls, payload: Dict[str, Any]) -> "JsonRpcResponse":
        if not (rpc_constants.JSON_RPC_RESULT not in payload) ^ (rpc_constants.JSON_RPC_ERROR not in payload):
            raise ValueError(
                "Cannot instantiate a message with neither (or both) a result and error."
            )
        return cls(
            payload.get(rpc_constants.JSON_RPC_REQUEST_ID, None),
            payload.get(rpc_constants.JSON_RPC_RESULT, None),
            utils.optional_map(
                payload.get(rpc_constants.JSON_RPC_ERROR, None),
                RpcError.from_json
            )
        )

    @classmethod
    def from_jsons(cls, payload: Union[bytes, str]) -> "JsonRpcResponse":
        return cls.from_json(json.loads(payload))



import time
from typing import Optional
import task_pool_executor as tpe

_executor: Optional[tpe.TaskPoolExecutor] = None


def init(thread_pool_parallelism_degree: int) -> None:
    executor = tpe.TaskPoolExecutor()
    executor.init(thread_pool_parallelism_degree)

    # pylint: disable=global-statement
    global _executor
    _executor = executor


# TODO : convert to async
def run_task(tsk: tpe.MainTaskBase) -> None:
    executor = _executor

    assert executor is not None
    executor.enqueue_task(tsk)
    while not tsk.is_completed():
        time.sleep(0)
        continue
    tsk.assert_execution()
    tsk.cleanup()


def get_pool_size() -> int:
    executor = _executor
    assert executor is not None
    return executor.size()



import os
import random
import typing
from abc import abstractmethod
from mock import MagicMock

from bxutils import logging
from bxutils.logging.log_level import LogLevel

from bxcommon.services.transaction_service import TransactionService
from bxcommon.test_utils import helpers
from bxcommon.utils import convert
from bxcommon.utils.blockchain_utils.ont.ont_object_hash import OntObjectHash

from bxgateway import ont_constants
from bxgateway.messages.ont.block_ont_message import BlockOntMessage
from bxgateway.messages.ont.inventory_ont_message import InventoryOntType
from bxgateway.services.ont.abstract_ont_block_cleanup_service import AbstractOntBlockCleanupService
from bxgateway.testing.abstract_block_cleanup_service_test import AbstractBlockCleanupServiceTest
from bxgateway.messages.ont import ont_messages_util

logger = logging.get_logger(__name__)


class AbstractOntBlockCleanupServiceTest(AbstractBlockCleanupServiceTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        logger.setLevel(LogLevel.INFO)

    def _get_sample_block(self, file_path):
        root_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(file_path))))
        with open(os.path.join(root_dir, "samples/ont_sample_block.txt")) as sample_file:
            ont_block = sample_file.read().strip("\n")
        buf = bytearray(convert.hex_to_bytes(ont_block))
        parsed_block = BlockOntMessage(buf=buf)
        return parsed_block

    def _test_mark_blocks_and_request_cleanup(self):
        self.node.has_active_blockchain_peer = MagicMock(return_value=True)
        self.cleanup_service._request_block = MagicMock()

        marked_block = OntObjectHash(binary=helpers.generate_bytearray(ont_constants.ONT_HASH_LEN))
        prev_block = OntObjectHash(binary=helpers.generate_bytearray(ont_constants.ONT_HASH_LEN))
        tracked_blocks = []
        self.cleanup_service.on_new_block_received(marked_block, prev_block)
        self.transaction_service.track_seen_short_ids(marked_block, [])
        for _ in range(self.block_confirmations_count - 1):
            tracked_block = OntObjectHash(binary=helpers.generate_bytearray(ont_constants.ONT_HASH_LEN))
            self.transaction_service.track_seen_short_ids(tracked_block, [])
            tracked_blocks.append(tracked_block)
        unmarked_block = OntObjectHash(binary=helpers.generate_bytearray(ont_constants.ONT_HASH_LEN))
        self.assertIsNone(self.cleanup_service.last_confirmed_block)
        self.cleanup_service.mark_blocks_and_request_cleanup([marked_block, *tracked_blocks])
        self.assertEqual(marked_block, self.cleanup_service.last_confirmed_block)
        self.assertTrue(self.cleanup_service.is_marked_for_cleanup(marked_block))
        self.assertFalse(self.cleanup_service.is_marked_for_cleanup(unmarked_block))
        self.assertEqual(marked_block, self.cleanup_service.last_confirmed_block)
        self.cleanup_service._request_block.assert_called_once()

    def _test_block_cleanup(self):
        block_msg = self._get_sample_block(self._get_file_path())
        transactions = block_msg.txns()
        block_hash = typing.cast(OntObjectHash, block_msg.block_hash())
        random.shuffle(transactions)
        short_len = int(len(transactions) * 0.9)
        transactions_short = transactions[:short_len]
        unknown_transactions = transactions[short_len:]
        transaction_keys = []
        for idx, tx in enumerate(transactions_short):
            tx_hash, _ = ont_messages_util.get_txid(tx)
            transaction_key = self.transaction_service.get_transaction_key(tx_hash)
            transaction_keys.append(transaction_key)
            self.transaction_service.set_transaction_contents_by_key(transaction_key, tx)
            self.transaction_service.assign_short_id_by_key(transaction_key, idx + 1)
        for idx, tx in enumerate(unknown_transactions):
            tx_hash, _= ont_messages_util.get_txid(tx)
            transaction_key = self.transaction_service.get_transaction_key(tx_hash)
            transaction_keys.append(transaction_key)
            if idx % 2 == 0:
                self.transaction_service.set_transaction_contents_by_key(transaction_key, tx)
        self.cleanup_service._block_hash_marked_for_cleanup.add(block_hash)
        self.cleanup_service.clean_block_transactions(block_msg, self.transaction_service)
        self.assertEqual(0, self.transaction_service._total_tx_contents_size)
        for transaction_key in transaction_keys:
            self.assertFalse(self.transaction_service.has_transaction_contents_by_key(transaction_key))

    @abstractmethod
    def _get_transaction_service(self) -> TransactionService:
        pass

    @abstractmethod
    def _get_cleanup_service(self) -> AbstractOntBlockCleanupService:
        pass

    @abstractmethod
    def _get_file_path(self) -> str:
        pass



from typing import TYPE_CHECKING, Optional

from bxcommon.models.blockchain_peer_info import BlockchainPeerInfo
from bxcommon.models.blockchain_protocol import BlockchainProtocol
from bxcommon.rpc.bx_json_rpc_request import BxJsonRpcRequest
from bxcommon.rpc.requests.abstract_rpc_request import AbstractRpcRequest
from bxcommon.rpc import rpc_constants
from bxcommon.rpc.rpc_errors import RpcInvalidParams, RpcInternalError
from bxgateway import argument_parsers

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode


class AbstractBlockchainPeerRpcRequest(AbstractRpcRequest["AbstractGatewayNode"]):
    help = {
        "params": f"{rpc_constants.BLOCKCHAIN_PEER_PARAMS_KEY}: "
                "For Ethereum, the format is enode://<eth node public key>@<eth node ip>:<port>. "
                "For other blockchain protocols, the format is <ip>:<port>",
    }

    def __init__(self, request: BxJsonRpcRequest, node: "AbstractGatewayNode") -> None:
        self._blockchain_peer_info: Optional[BlockchainPeerInfo] = None
        super().__init__(request, node)

    def validate_params(self) -> None:
        super().validate_params()
        params = self.params
        if params is None or not isinstance(params, dict):
            raise RpcInvalidParams(
                self.request_id,
                "Params request field is either missing or not a dictionary type."
            )
        if rpc_constants.BLOCKCHAIN_PEER_PARAMS_KEY in params:
            peer = params[rpc_constants.BLOCKCHAIN_PEER_PARAMS_KEY]
            blockchain_protocol = self.node.opts.blockchain_protocol
            if blockchain_protocol is not None:
                self._blockchain_peer_info = self.parse_peer(blockchain_protocol, peer)
            else:
                raise RpcInternalError(
                    self.request_id,
                    "Could not process request to add/remove blockchain peer. Please contact bloXroute support."
                )
        else:
            raise RpcInvalidParams(
                self.request_id,
                f"Missing param: {rpc_constants.BLOCKCHAIN_PEER_PARAMS_KEY}."
            )

    def parse_enode(self, enode: str) -> BlockchainPeerInfo:
        # Make sure enode is at least as long as the public key
        if not argument_parsers.enode_is_valid_length(enode):
            raise RpcInvalidParams(
                self.request_id,
                f"Invalid enode: {enode}, with length: {len(enode)}. "
                f"Expected format: enode://<eth node public key>@<eth node ip>:<port>"
            )
        try:
            pub_key, ip, port = argument_parsers.get_enode_parts(enode)
            if not port.isnumeric():
                raise RpcInvalidParams(
                    self.request_id,
                    f"Invalid port: {port}"
                )
        except ValueError:
            raise RpcInvalidParams(
                self.request_id,
                f"Invalid enode: {enode}. "
                f"Expected format: enode://<eth node public key>@<eth node ip>:<port>"
            )
        else:
            return BlockchainPeerInfo(ip, int(port), pub_key)

    def parse_ip_port(self, ip_port_string: str) -> BlockchainPeerInfo:
        ip, port = argument_parsers.get_ip_port_string_parts(ip_port_string)
        if not port.isnumeric():
            raise RpcInvalidParams(
                self.request_id,
                f"Invalid port: {port}"
            )
        return BlockchainPeerInfo(ip, int(port))

    def parse_peer(self, blockchain_protocol: str, peer: str) -> BlockchainPeerInfo:
        if blockchain_protocol.lower() == BlockchainProtocol.ETHEREUM.value:
            return self.parse_enode(peer)
        else:
            return self.parse_ip_port(peer)



from typing import Union, Dict, Callable

from bxcommon.utils.blockchain_utils.eth import transaction_validation_utils
from bxcommon.models.blockchain_protocol import BlockchainProtocol
from bxcommon.models.tx_validation_status import TxValidationStatus

protocol_transaction_validation: Dict[BlockchainProtocol, Callable[
    [Union[bytes, bytearray, memoryview], int],
    TxValidationStatus
]]
network_num_to_protocol: Dict[int, str]


def validate_transaction(
    tx_bytes: Union[bytes, bytearray, memoryview],
    protocol: BlockchainProtocol,
    min_tx_network_fee: int
) -> TxValidationStatus:
    return protocol_transaction_validation[protocol](tx_bytes, min_tx_network_fee)


def btc_validate_transaction(
    _tx_bytes: Union[bytes, bytearray, memoryview],
    _min_tx_network_fee: int
) -> TxValidationStatus:
    return TxValidationStatus.VALID_TX


def eth_validate_transaction(
    tx_bytes: Union[bytes, bytearray, memoryview],
    min_tx_network_fee: int
) -> TxValidationStatus:
    if isinstance(tx_bytes, bytes):
        tx_bytes = bytearray(tx_bytes)
    return transaction_validation_utils.validate_transaction(tx_bytes, min_tx_network_fee)


def ont_validate_transaction(
    _tx_bytes: Union[bytes, bytearray, memoryview],
    _min_tx_network_fee: int
) -> TxValidationStatus:
    return TxValidationStatus.VALID_TX


protocol_transaction_validation = {
    BlockchainProtocol.BITCOIN: btc_validate_transaction,
    BlockchainProtocol.BITCOINCASH: btc_validate_transaction,
    BlockchainProtocol.ETHEREUM: eth_validate_transaction,
    BlockchainProtocol.ONTOLOGY: ont_validate_transaction

}



import time
from typing import TYPE_CHECKING, Iterable

import task_pool_executor as tpe

from bxcommon.messages.bloxroute.block_confirmation_message import BlockConfirmationMessage
from bxcommon.services import extension_cleanup_service_helpers
from bxcommon.services.transaction_service import TransactionService
from bxcommon.utils.object_hash import Sha256Hash
from bxcommon.utils.proxy.task_queue_proxy import TaskQueueProxy
from bxgateway.services.eth.abstract_eth_block_cleanup_service import AbstractEthBlockCleanupService
from bxutils import logging
from bxutils.logging.log_record_type import LogRecordType


if TYPE_CHECKING:
    from bxgateway.connections.eth.eth_gateway_node import EthGatewayNode

logger = logging.get_logger(LogRecordType.BlockCleanup, __name__)


def create_cleanup_task() -> tpe.BlockConfirmationCleanupTask:
    return tpe.BlockConfirmationCleanupTask()


class EthExtensionBlockCleanupService(AbstractEthBlockCleanupService):
    """
    Service for managing block cleanup.
    """

    def __init__(self, node: "EthGatewayNode", network_num: int):
        super().__init__(node, network_num)
        self.cleanup_tasks = TaskQueueProxy(create_cleanup_task)

    def clean_block_transactions_by_block_components(
            self,
            block_hash: Sha256Hash,
            transactions_list: Iterable[Sha256Hash],
            transaction_service: TransactionService
         ) -> None:
        logger.debug("Processing block for cleanup: {}", block_hash)
        tx_hash_to_contents_len_before_cleanup = transaction_service.get_tx_hash_to_contents_len()
        short_id_count_before_cleanup = transaction_service.get_short_id_count()

        start_time = time.time()
        if not isinstance(transactions_list, list):
            transactions_list = list(transactions_list)
        sids = []
        confirmation_msg = BlockConfirmationMessage(
            block_hash, transaction_service.network_num, tx_hashes=transactions_list, sids=sids
        )
        self.contents_cleanup(transaction_service, confirmation_msg)
        self._block_hash_marked_for_cleanup.discard(block_hash)
        end_time = time.time()
        duration = end_time - start_time
        logger.debug(
            "Finished cleaning up block {}. Processed {} hashes. Took {:.3f}s.",
            block_hash,
            len(transactions_list),
            duration
        )
        self.node.post_block_cleanup_tasks(
            block_hash,
            sids,
            transactions_list
        )

        transactions_processed = len(transactions_list)
        tx_hash_to_contents_len_after_cleanup = transaction_service.get_tx_hash_to_contents_len()
        short_id_count_after_cleanup = transaction_service.get_short_id_count()

        transaction_service.log_block_transaction_cleanup_stats(block_hash, transactions_processed,
                                                                tx_hash_to_contents_len_before_cleanup,
                                                                tx_hash_to_contents_len_after_cleanup,
                                                                short_id_count_before_cleanup,
                                                                short_id_count_after_cleanup)

    # pyre-fixme[14]: `contents_cleanup` overrides method defined in
    #  `AbstractBlockCleanupService` inconsistently.
    def contents_cleanup(
            self,
            transaction_service: TransactionService,
            block_confirmation_message: BlockConfirmationMessage
     ):
        extension_cleanup_service_helpers.contents_cleanup(
            transaction_service, block_confirmation_message, self.cleanup_tasks
        )



# pyre-ignore-all-errors
# This file does type manipulation that type checkers will not understand.

import dataclasses
import json
from enum import Flag
from typing import Dict, Type, TypeVar, Any, List, Optional, Union, Set

import humps

from bxcommon.utils import convert
from bxutils import log_messages
from bxutils import logging
from bxutils.encoding.json_encoder import Case

logger = logging.get_logger(__name__)
T = TypeVar("T")
_TYPE_HANDLER_MAPPING = {}
STRING_PARSING_METHOD = "from_string"
DICT_PARSING_METHOD = "from_json"


def load_model_from_json(
    model_class: Type[T], model_params: str, model_case: Case = Case.SNAKE
) -> T:
    try:
        model_dict = json.loads(model_params)
        return load_model(model_class, model_dict, model_case)
    except Exception:
        logger.error(log_messages.ERROR_LOADING_MODEL_INTO_DICT, model_class, model_params)
        raise


def load_model(
    model_class: Type[T],
    model_params: Dict[str, Any],
    model_case: Case = Case.SNAKE
) -> T:
    """
    Loads a model class instance from the dictionary.

    This function ensures that models are forward compatible. If attributes are
    added to models in future versions, and those versions are serialized somewhere
    (e.g. Redis), this function ensures that only attributes that the current
    version knows about are loaded.

    :param model_class: class of the model to load into
    :param model_params: attributes from the retrieved model to load
    :param model_case: case of the attribute names from the retrieved model
                       these attribute names will be converted to snake_case for lookup
                       of type information
    :return: instance of model class
    """
    try:
        attributes = {}
        for attribute in dataclasses.fields(model_class):
            if model_case == Case.SNAKE:
                model_attribute_name = attribute.name
            else:
                model_attribute_name = humps.camelize(attribute.name)
            if model_attribute_name in model_params:
                attributes[attribute.name] = _load_attribute(
                    attribute.type,
                    model_params[model_attribute_name],
                    model_case=model_case
                )
        return model_class(**attributes)
    except Exception as e:
        raise TypeError(
            f"Could not load model of type {model_class} from data: {model_params}"
        ) from e


# pylint: disable=too-many-return-statements,too-many-branches
def _load_attribute(
    attribute_type: Type[T],
    attribute_value: Any,
    cast_basic_values: bool = True,
    model_case: Case = Case.SNAKE,
) -> Optional[T]:
    """
    Loads an attribute and attempts to cast it to the correct class type.

    For any type with a custom `from_string` method, attempt to construct using that method.
    For data classes, recurse and attempt to load the original model.
    For Any types, None values, or Generic types, return the provided value as is.
    For built in `typing` types, special handling is generally required. List/Dict/... have been implemented for now.
    For Flag enum types, use the Flag name for loading / writing.
    For basic types, construct the objects normally.
    :param attribute_type: class of the attribute to load
    :param attribute_value: attribute value to parse
    :param cast_basic_values: if to cast between str, int, etc.
    :return: instance of attribute value
    """
    if (
        isinstance(attribute_value, str)
        and hasattr(attribute_type, STRING_PARSING_METHOD)
        and callable(getattr(attribute_type, STRING_PARSING_METHOD))
    ):
        return getattr(attribute_type, STRING_PARSING_METHOD)(attribute_value)
    elif (
        isinstance(attribute_value, dict)
        and hasattr(attribute_type, DICT_PARSING_METHOD)
        and callable(getattr(attribute_type, DICT_PARSING_METHOD))
    ):
        return getattr(attribute_type, DICT_PARSING_METHOD)(attribute_value)
    elif hasattr(attribute_type, "__annotations__"):
        return load_model(attribute_type, attribute_value, model_case)
    elif attribute_type == Any or attribute_value is None or attribute_type.__class__ == TypeVar:
        return attribute_value
    elif hasattr(attribute_type, "__origin__"):
        if attribute_type.__origin__ in _TYPE_HANDLER_MAPPING:
            return _TYPE_HANDLER_MAPPING[attribute_type.__origin__](
                attribute_type, attribute_value, model_case
            )
        else:
            raise NotImplementedError(
                f"Model loader is not capable of loading a "
                f"{attribute_type.__origin__} type. Please implement a handler."
            )
    elif issubclass(attribute_type, Flag):
        return attribute_type[str(attribute_value)]
    else:
        if cast_basic_values:
            try:
                return attribute_type(attribute_value)
            except Exception as e:
                # backup processing for hex string types
                try:
                    if attribute_type == int:
                        return int(attribute_value, 16)
                    if attribute_type == bytearray:
                        return convert.hex_to_bytes(attribute_value)
                except Exception as _e:
                    raise e
                raise e
        elif not isinstance(attribute_value, attribute_type):
            raise TypeError(f"{attribute_value} is not of type {attribute_type}.")
        else:
            return attribute_value


def _load_list(list_type: Type[T], list_value: Any, model_case: Case) -> List:
    if isinstance(list_value, dict):
        raise ValueError("Forcibly raising when trying to cast Dict to List.")

    if len(list_type.__args__) != 1:
        raise ValueError("List type annotation requires 1 or 0 args.")

    list_param = list_type.__args__[0]
    return [
        _load_attribute(list_param, list_entry, model_case=model_case)
        for list_entry in list_value
    ]


def _load_set(set_type: Type[T], set_value: Any, model_case: Case) -> Set:
    if isinstance(set_value, dict):
        raise ValueError("Forcibly raising when trying to cast Dict to Set.")

    if len(set_type.__args__) != 1:
        raise ValueError("List type annotation requires 1 or 0 args.")

    set_param = set_type.__args__[0]
    return {
        _load_attribute(set_param, set_entry, model_case=model_case)
        for set_entry in set_value
    }


def _load_dict(dict_type: Type[T], dict_value: Any, model_case: Case) -> Dict:
    if not isinstance(dict_value, Dict):
        raise TypeError("Cannot deserialize value {} to Dict type {}".format(dict_value, dict_type))

    if len(dict_type.__args__) != 2:
        raise ValueError("Dict type annotation requires 2 or 0 args.")

    key_type = dict_type.__args__[0]
    value_type = dict_type.__args__[1]
    return {
        _load_attribute(
            key_type, key, model_case=model_case
        ): _load_attribute(value_type, value, model_case=model_case)
        for key, value in dict_value.items()
    }


def _load_optional(optional_type: Type[T], optional_value: Any, model_case: Case) -> Optional:
    if len(optional_type.__args__) != 1:
        raise ValueError("Optional types must have a type annotation.")

    optional_param = optional_type.__args__[0]
    return _load_attribute(optional_param, optional_value, model_case=model_case)


def _load_union(union_type: Type[T], union_value: Any, model_case: Case) -> Any:
    # pylint: disable=unidiomatic-typecheck
    if len(union_type.__args__) == 2 and type(None) in union_type.__args__:
        # special case of an optional
        return _load_attribute(union_type.__args__[0], union_value, model_case=model_case)
    else:
        for union_param in union_type.__args__:
            # noinspection PyBroadException
            try:
                return _load_attribute(
                    union_param, union_value, cast_basic_values=False, model_case=model_case
                )
            # pylint: disable=broad-except
            except Exception as _e:
                continue
        raise NotImplementedError(
            f"Could not find a union type that matched value. "
            f"Types: {union_type.__args__}, Value: {union_value}"
        )


_TYPE_HANDLER_MAPPING.update({
    list: _load_list,
    List: _load_list,
    set: _load_set,
    Set: _load_set,
    dict: _load_dict,
    Dict: _load_dict,
    Optional: _load_optional,
    Union: _load_union,
})



from typing import Union, Tuple

from bxcommon.utils import crypto
from bxcommon.utils.blockchain_utils.btc.btc_common_utils import btc_varint_to_int
from bxcommon.utils.blockchain_utils.ont import ont_common_constants
from bxcommon.utils.blockchain_utils.ont.ont_object_hash import OntObjectHash


def ont_varint_to_int(buf: Union[memoryview, bytearray], off: int) -> Tuple[int, int]:
    return btc_varint_to_int(buf, off)


def get_txid(buffer: Union[memoryview, bytearray]) -> Tuple[OntObjectHash, int]:
    if not isinstance(buffer, memoryview):
        buffer = memoryview(buffer)
    off = 1
    txtype = buffer[off:off + 1]
    off += 1 + 40
    # Deploy type
    if txtype == ont_common_constants.ONT_TX_DEPLOY_TYPE_INDICATOR_AS_BYTEARRAY:
        buffer_length, size = ont_varint_to_int(buffer, off)
        off += buffer_length + size
        off += 1
        buffer_length, size = ont_varint_to_int(buffer, off)
        off += buffer_length + size
        buffer_length, size = ont_varint_to_int(buffer, off)
        off += buffer_length + size
        buffer_length, size = ont_varint_to_int(buffer, off)
        off += buffer_length + size
        buffer_length, size = ont_varint_to_int(buffer, off)
        off += buffer_length + size
        buffer_length, size = ont_varint_to_int(buffer, off)
        off += buffer_length + size
    # Invoke type
    elif txtype == ont_common_constants.ONT_TX_INVOKE_TYPE_INDICATOR_AS_BYTEARRAY:
        buffer_length, size = ont_varint_to_int(buffer, off)
        off += buffer_length + size

    _, size = ont_varint_to_int(buffer, off)
    off += size

    return OntObjectHash(buf=crypto.double_sha256(buffer[:off]), length=ont_common_constants.ONT_HASH_LEN), off



import time
from collections import deque
from typing import List, Deque, cast, Union

from bxcommon.feed.feed import FeedKey
from bxcommon.messages.abstract_message import AbstractMessage
from bxcommon.utils import convert
from bxcommon.utils.blockchain_utils.eth import eth_common_utils, eth_common_constants
from bxcommon.utils.expiring_dict import ExpiringDict
from bxcommon.utils.object_hash import Sha256Hash, NULL_SHA256_HASH
from bxcommon.utils.stats.block_stat_event_type import BlockStatEventType
from bxcommon.utils.stats.block_statistics_service import block_stats
from bxcommon.feed.feed_source import FeedSource
from bxgateway import log_messages
from bxgateway.connections.eth.eth_base_connection_protocol import EthBaseConnectionProtocol
from bxgateway.eth_exceptions import CipherNotInitializedError
from bxcommon.feed.eth.eth_new_transaction_feed import EthNewTransactionFeed
from bxcommon.feed.eth.eth_pending_transaction_feed import EthPendingTransactionFeed
from bxcommon.feed.eth.eth_raw_transaction import EthRawTransaction
from bxgateway.messages.eth.internal_eth_block_info import InternalEthBlockInfo
from bxgateway.messages.eth.new_block_parts import NewBlockParts
from bxgateway.messages.eth.protocol.block_bodies_eth_protocol_message import BlockBodiesEthProtocolMessage
from bxgateway.messages.eth.protocol.block_headers_eth_protocol_message import BlockHeadersEthProtocolMessage
from bxgateway.messages.eth.protocol.eth_protocol_message_type import EthProtocolMessageType
from bxgateway.messages.eth.protocol.get_block_bodies_eth_protocol_message import GetBlockBodiesEthProtocolMessage
from bxgateway.messages.eth.protocol.get_block_headers_eth_protocol_message import GetBlockHeadersEthProtocolMessage
from bxgateway.messages.eth.protocol.get_node_data_eth_protocol_message import GetNodeDataEthProtocolMessage
from bxgateway.messages.eth.protocol.get_pooled_transactions_eth_protocol_message import \
    GetPooledTransactionsEthProtocolMessage
from bxgateway.messages.eth.protocol.get_receipts_eth_protocol_message import GetReceiptsEthProtocolMessage
from bxgateway.messages.eth.protocol.new_block_eth_protocol_message import NewBlockEthProtocolMessage
from bxgateway.messages.eth.protocol.new_block_hashes_eth_protocol_message import NewBlockHashesEthProtocolMessage
from bxgateway.messages.eth.protocol.new_pooled_transaction_hashes_eth_protocol_message import \
    NewPooledTransactionHashesEthProtocolMessage
from bxgateway.messages.eth.protocol.status_eth_protocol_message import StatusEthProtocolMessage
from bxgateway.messages.eth.protocol.status_eth_protocol_message_v63 import StatusEthProtocolMessageV63
from bxgateway.messages.eth.protocol.transactions_eth_protocol_message import \
    TransactionsEthProtocolMessage
from bxgateway.services.eth.eth_block_queuing_service import EthBlockQueuingService
from bxgateway.services.gateway_transaction_service import ProcessTransactionMessageFromNodeResult
from bxgateway.utils.eth.rlpx_cipher import RLPxCipher
from bxgateway.utils.stats.gateway_bdn_performance_stats_service import \
    gateway_bdn_performance_stats_service
from bxgateway.utils.stats.gateway_transaction_stats_service import gateway_transaction_stats_service
from bxgateway.utils.stats.transaction_feed_stats_service import transaction_feed_stats_service
from bxutils import logging

logger = logging.get_logger(__name__)


class EthNodeConnectionProtocol(EthBaseConnectionProtocol):
    def __init__(self, connection, is_handshake_initiator, rlpx_cipher: RLPxCipher):
        super().__init__(connection, is_handshake_initiator, rlpx_cipher)
        #  NOTE - cannot call `msg_proxy_request` directly
        connection.message_handlers.update({
            EthProtocolMessageType.STATUS: self.msg_status,
            EthProtocolMessageType.TRANSACTIONS: self.msg_tx,
            EthProtocolMessageType.NEW_POOLED_TRANSACTION_HASHES: self.msg_new_pooled_tx_hashes,
            EthProtocolMessageType.POOLED_TRANSACTIONS: self.msg_tx,
            EthProtocolMessageType.GET_BLOCK_HEADERS: self.msg_get_block_headers,
            EthProtocolMessageType.GET_BLOCK_BODIES: self.msg_get_block_bodies,
            EthProtocolMessageType.GET_NODE_DATA: self.msg_get_node_data,
            EthProtocolMessageType.GET_RECEIPTS: self.msg_get_receipts,
            EthProtocolMessageType.BLOCK_HEADERS: self.msg_block_headers,
            EthProtocolMessageType.NEW_BLOCK: self.msg_block,
            EthProtocolMessageType.NEW_BLOCK_HASHES: self.msg_new_block_hashes,
            EthProtocolMessageType.BLOCK_BODIES: self.msg_block_bodies
        })

        self.waiting_checkpoint_headers_request = True

        # uses block hash as a key, and NewBlockParts structure as value
        self.pending_new_block_parts: ExpiringDict[
            Sha256Hash, NewBlockParts
        ] = ExpiringDict(
            self.node.alarm_queue,
            eth_common_constants.NEW_BLOCK_PARTS_MAX_WAIT_S,
            f"{str(self)}_eth_pending_block_parts"
        )

        # hashes of new blocks constructed from headers and bodies, ready to send to BDN
        self._ready_new_blocks: Deque[Sha256Hash] = deque()

        # queue of lists of hashes that are awaiting block bodies response
        self._block_bodies_requests: Deque[List[Sha256Hash]] = deque(
            maxlen=eth_common_constants.REQUESTED_NEW_BLOCK_BODIES_MAX_COUNT)

        if self.block_cleanup_poll_interval_s > 0:
            self.connection.node.alarm_queue.register_alarm(
                self.block_cleanup_poll_interval_s,
                self._request_blocks_confirmation
            )

        self.requested_blocks_for_confirmation: ExpiringDict[
            Sha256Hash, float
        ] = ExpiringDict(
            self.node.alarm_queue,
            eth_common_constants.BLOCK_CONFIRMATION_REQUEST_CACHE_INTERVAL_S,
            f"{str(self)}_eth_requested_blocks"
        )
        self._connection_established_time = 0.0

    def msg_status(self, msg: Union[StatusEthProtocolMessage, StatusEthProtocolMessageV63]):
        super(EthNodeConnectionProtocol, self).msg_status(msg)

        self.connection.on_connection_established()

        self.connection.schedule_pings()

        self.node.on_blockchain_connection_ready(self.connection)
        self.node.alarm_queue.register_alarm(
            eth_common_constants.CHECKPOINT_BLOCK_HEADERS_REQUEST_WAIT_TIME_S,
            self._stop_waiting_checkpoint_headers_request
        )

        self._connection_established_time = time.time()

    def msg_new_pooled_tx_hashes(self, msg: NewPooledTransactionHashesEthProtocolMessage) -> None:
        transaction_hashes = msg.transaction_hashes()

        if not self.node.opts.has_fully_updated_tx_service:
            self.connection.log_debug(
                "Skipping {} announced transactions from Ethereum peer, not yet synced", len(transaction_hashes)
            )
            return

        new_tx_hashes = []
        for tx_hash in transaction_hashes:
            if not self.tx_service.has_transaction_contents_by_key(
                self.tx_service.get_transaction_key(tx_hash, None)
            ):
                new_tx_hashes.append(tx_hash.binary)

        if new_tx_hashes:
            self.connection.enqueue_msg(
                GetPooledTransactionsEthProtocolMessage(
                    None,
                    new_tx_hashes
                )
            )
            self.connection.log_trace("Fetching {} announced transactions from Ethereum peer", len(new_tx_hashes))

    def msg_tx(self, msg: TransactionsEthProtocolMessage) -> None:
        if len(msg.rawbytes()) >= eth_common_constants.ETH_SKIP_TRANSACTIONS_SIZE:
            bytes_skipped_count = len(msg.rawbytes())
            self.connection.log_debug("Skipping {} bytes of transactions message.", bytes_skipped_count)
            gateway_transaction_stats_service.log_skipped_transaction_bytes(bytes_skipped_count)
        else:
            super().msg_tx(msg)

    def msg_tx_after_tx_service_process_complete(self, process_result: List[ProcessTransactionMessageFromNodeResult]):
        # calculate minimal tx gas price only if transaction validation is enabled
        if not self.node.opts.transaction_validation:
            return

        for result in process_result:
            fee = eth_common_utils.raw_tx_gas_price(memoryview(result.transaction_contents), 0)
            self.node.min_tx_from_node_gas_price.add(fee)

    # pyre-fixme[14]: `msg_block` overrides method defined in
    #  `AbstractBlockchainConnectionProtocol` inconsistently.
    def msg_block(self, msg: NewBlockEthProtocolMessage) -> None:
        if not self.node.should_process_block_hash(msg.block_hash()):
            return

        self.node.set_known_total_difficulty(msg.block_hash(), msg.get_chain_difficulty())

        internal_new_block_msg = InternalEthBlockInfo.from_new_block_msg(msg)
        self.process_msg_block(internal_new_block_msg, msg.number())

        if self.node.opts.filter_txs_factor > 0:
            self.node.on_transactions_in_block(msg.txns())

    def msg_new_block_hashes(self, msg: NewBlockHashesEthProtocolMessage):
        if not self.node.should_process_block_hash(msg.block_hash()):
            return

        block_hash_number_pairs = []
        for block_hash, block_number in msg.get_block_hash_number_pairs():
            block_stats.add_block_event_by_block_hash(
                block_hash,
                BlockStatEventType.BLOCK_ANNOUNCED_BY_BLOCKCHAIN_NODE,
                network_num=self.connection.network_num,
                more_info="Protocol: {}, Network: {}. {}".format(
                    self.node.opts.blockchain_protocol,
                    self.node.opts.blockchain_network,
                    msg.extra_stats_data()
                ),
                block_height=block_number,
            )
            gateway_bdn_performance_stats_service.log_block_message_from_blockchain_node(
                self.connection.endpoint,
                False
            )

            if block_hash in self.node.blocks_seen.contents:
                self.node.on_block_seen_by_blockchain_node(block_hash, self.connection, block_number=block_number)
                block_stats.add_block_event_by_block_hash(
                    block_hash,
                    BlockStatEventType.BLOCK_RECEIVED_FROM_BLOCKCHAIN_NODE_IGNORE_SEEN,
                    network_num=self.connection.network_num,
                    block_height=block_number,
                )
                self.connection.log_info(
                    "Ignoring duplicate block {} from local blockchain node.",
                    block_hash
                )
                continue

            self.node.on_block_seen_by_blockchain_node(block_hash, self.connection, block_number=block_number)

            self.node.track_block_from_node_handling_started(block_hash)
            block_hash_number_pairs.append((block_hash, block_number))

            self.connection.log_info(
                "Fetching block {} from local Ethereum node.",
                block_hash
            )

        if not block_hash_number_pairs:
            return

        for block_hash, block_number in block_hash_number_pairs:
            # pyre-fixme[6]: Expected `memoryview` for 1st param but got `None`.
            self.pending_new_block_parts.add(block_hash, NewBlockParts(None, None, block_number))
            self.connection.enqueue_msg(GetBlockHeadersEthProtocolMessage(None, block_hash.binary, 1, 0, 0))

        self.request_block_body([block_hash for block_hash, _ in block_hash_number_pairs])

    def request_block_body(self, block_hashes: List[Sha256Hash]):
        block_request_message = GetBlockBodiesEthProtocolMessage(
            None,
            block_hashes=[bytes(blk.binary) for blk in block_hashes]
        )
        self.connection.enqueue_msg(block_request_message)
        self._block_bodies_requests.append(block_hashes)

    def msg_get_block_headers(self, msg: GetBlockHeadersEthProtocolMessage):
        if self._waiting_checkpoint_headers_request:
            super(EthNodeConnectionProtocol, self).msg_get_block_headers(msg)
        else:
            block_queuing_service = cast(
                EthBlockQueuingService,
                self.node.block_queuing_service_manager.get_block_queuing_service(self.connection)
            )
            processed = self.node.block_processing_service.try_process_get_block_headers_request(
                msg,
                block_queuing_service
            )
            if not processed:
                self.msg_proxy_request(msg, self.connection)

    def msg_get_block_bodies(self, msg: GetBlockBodiesEthProtocolMessage):
        block_queuing_service = cast(
            EthBlockQueuingService,
            self.node.block_queuing_service_manager.get_block_queuing_service(self.connection)
        )
        processed = self.node.block_processing_service.try_process_get_block_bodies_request(msg, block_queuing_service)

        if not processed:
            self.node.log_requested_remote_blocks(msg.get_block_hashes())
            self.msg_proxy_request(msg, self.connection)

    def msg_block_headers(self, msg: BlockHeadersEthProtocolMessage):
        if not self.node.should_process_block_hash():
            return

        block_headers = msg.get_block_headers()

        if self.pending_new_block_parts.contents and len(block_headers) == 1:
            header_bytes = msg.get_block_headers_bytes()[0]
            block_hash_bytes = eth_common_utils.keccak_hash(header_bytes)
            block_hash = Sha256Hash(block_hash_bytes)

            if block_hash in self.pending_new_block_parts.contents:
                logger.debug("Received block header for new block {}", convert.bytes_to_hex(block_hash.binary))
                self.pending_new_block_parts.contents[block_hash].block_header_bytes = header_bytes
                self._check_pending_new_block(block_hash)
                self._process_ready_new_blocks()
                return

        if len(block_headers) > 0:
            block_hashes = [blk.hash_object() for blk in block_headers]
            block_hashes.insert(0, Sha256Hash(block_headers[0].prev_hash))
            self.node.block_cleanup_service.mark_blocks_and_request_cleanup(block_hashes)

        latest_block_number = 0
        latest_block_difficulty = 0

        block_queuing_service = self.node.block_queuing_service_manager.get_block_queuing_service(self.connection)
        if block_queuing_service is not None:
            for block_header in block_headers:
                block_queuing_service.mark_block_seen_by_blockchain_node(
                    block_header.hash_object(), None, block_header.number
                )

                if block_header.number > latest_block_number:
                    latest_block_number = block_header.number
                    latest_block_difficulty = block_header.difficulty

        self.node.block_processing_service.set_last_confirmed_block_parameters(
            latest_block_number, latest_block_difficulty
        )

    def msg_block_bodies(self, msg: BlockBodiesEthProtocolMessage):
        if not self.node.should_process_block_hash():
            return

        if self._block_bodies_requests:
            requested_hashes = self._block_bodies_requests.popleft()

            bodies_bytes = msg.get_block_bodies_bytes()

            if len(requested_hashes) != len(bodies_bytes):
                logger.debug("Expected {} bodies in response but received {}. Proxy message to remote node.",
                             len(requested_hashes), len(bodies_bytes))
                self._block_bodies_requests.clear()
                return

            logger.debug("Processing expected block bodies messages for blocks [{}]",
                         ", ".join([convert.bytes_to_hex(block_hash.binary) for block_hash in requested_hashes]))

            for block_hash, block_body_bytes in zip(requested_hashes, bodies_bytes):
                if block_hash in self.pending_new_block_parts.contents:
                    logger.debug("Received block body for pending new block {}",
                                 convert.bytes_to_hex(block_hash.binary))
                    self.pending_new_block_parts.contents[block_hash].block_body_bytes = block_body_bytes
                    self._check_pending_new_block(block_hash)
                elif self.node.block_cleanup_service.is_marked_for_cleanup(block_hash):
                    transactions_hashes = \
                        BlockBodiesEthProtocolMessage.from_body_bytes(block_body_bytes).get_block_transaction_hashes(0)
                    # pyre-fixme[16]: `AbstractBlockCleanupService` has no attribute
                    #  `clean_block_transactions_by_block_components`.
                    self.node.block_cleanup_service.clean_block_transactions_by_block_components(
                        transaction_service=self.node.get_tx_service(),
                        block_hash=block_hash,
                        transactions_list=transactions_hashes
                    )
                else:
                    logger.warning(log_messages.REDUNDANT_BLOCK_BODY,
                                   convert.bytes_to_hex(block_hash.binary))

            self._process_ready_new_blocks()

    def msg_get_receipts(self, msg: GetReceiptsEthProtocolMessage) -> None:
        self.node.log_requested_remote_blocks(msg.get_block_hashes())
        self.msg_proxy_request(msg, self.connection)

    def msg_get_node_data(self, msg: GetNodeDataEthProtocolMessage) -> None:
        self.msg_proxy_request(msg, self.connection)

    def _stop_waiting_checkpoint_headers_request(self):
        self._waiting_checkpoint_headers_request = False

    def _check_pending_new_block(self, block_hash):
        if block_hash in self.pending_new_block_parts.contents:
            pending_new_block = self.pending_new_block_parts.contents[block_hash]

            if pending_new_block.block_header_bytes is not None and pending_new_block.block_body_bytes is not None:
                self._ready_new_blocks.append(block_hash)

    def _process_ready_new_blocks(self):
        while self._ready_new_blocks:
            ready_block_hash = self._ready_new_blocks.pop()
            pending_new_block = self.pending_new_block_parts.contents[ready_block_hash]
            self.pending_new_block_parts.remove_item(ready_block_hash)

            if ready_block_hash in self.node.blocks_seen.contents:
                self.node.on_block_seen_by_blockchain_node(
                    ready_block_hash, self.connection, block_number=pending_new_block.block_number
                )
                self.connection.log_info(
                    "Discarding already seen block {} received in block bodies msg from local blockchain node.",
                    ready_block_hash
                )
                return

            last_known_total_difficulty = self.node.try_calculate_total_difficulty(
                ready_block_hash,
                pending_new_block
            )
            new_block_msg = InternalEthBlockInfo.from_new_block_parts(
                pending_new_block,
                last_known_total_difficulty
            )

            if self.is_valid_block_timestamp(new_block_msg):
                block_stats.add_block_event_by_block_hash(
                    ready_block_hash,
                    BlockStatEventType.BLOCK_RECEIVED_FROM_BLOCKCHAIN_NODE,
                    network_num=self.connection.network_num,
                    more_info="Protocol: {}, Network: {}. {}".format(
                        self.node.opts.blockchain_protocol,
                        self.node.opts.blockchain_network,
                        new_block_msg.extra_stats_data()
                    ),
                    block_height=new_block_msg.block_number(),
                )
                gateway_bdn_performance_stats_service.log_block_from_blockchain_node(self.connection.endpoint)

                canceled_recovery = self.node.on_block_seen_by_blockchain_node(
                    ready_block_hash, self.connection, new_block_msg, pending_new_block.block_number
                )
                if canceled_recovery:
                    return

                self.node.block_queuing_service_manager.push(
                    ready_block_hash, new_block_msg, node_received_from=self.connection
                )
                self.node.block_processing_service.queue_block_for_processing(
                    new_block_msg, self.connection
                )

    def _request_blocks_confirmation(self):
        # TODO: remove unused method
        try:
            super(EthNodeConnectionProtocol, self)._request_blocks_confirmation()
        except CipherNotInitializedError:
            logger.info(
                "Failed to request block confirmation due to bad cipher state, "
                "probably because the handshake was not completed yet."
            )
        return self.block_cleanup_poll_interval_s

    def _build_get_blocks_message_for_block_confirmation(self, hashes: List[Sha256Hash]) -> AbstractMessage:
        # TODO: remove unused method
        block_hash = NULL_SHA256_HASH
        for block_hash in hashes:
            last_attempt = self.requested_blocks_for_confirmation.contents.get(block_hash, 0)
            if time() - last_attempt > \
                self.block_cleanup_poll_interval_s * eth_common_constants.BLOCK_CONFIRMATION_REQUEST_INTERVALS:
                break
        else:
            block_hash = hashes[0]

        self.requested_blocks_for_confirmation.add(block_hash, time.time())
        return GetBlockHeadersEthProtocolMessage(
            None,
            block_hash=bytes(block_hash.binary),
            amount=100,
            skip=0,
            reverse=0
        )

    def publish_transaction(
        self, tx_hash: Sha256Hash, tx_contents: memoryview, local_region: bool = True
    ) -> None:
        transaction_feed_stats_service.log_new_transaction(tx_hash)
        self.node.feed_manager.publish_to_feed(
            FeedKey(EthNewTransactionFeed.NAME, self.node.network_num),
            EthRawTransaction(
                tx_hash,
                tx_contents,
                FeedSource.BLOCKCHAIN_SOCKET,
                local_region=local_region
            )
        )
        transaction_feed_stats_service.log_pending_transaction_from_local(tx_hash)
        self.node.feed_manager.publish_to_feed(
            FeedKey(EthPendingTransactionFeed.NAME, network_num=self.node.network_num),
            EthRawTransaction(
                tx_hash,
                tx_contents,
                FeedSource.BLOCKCHAIN_SOCKET,
                local_region=local_region
            )
        )
